#include "lsqrms.h"
#include "umath.h"
#include "matvecalloc.h"
#include "matvecutils.h"

static int solve(LDBL [], LDBL, LDBL, LDBL, LDBL);
static double TRvec(double *, double **, double *);
static LDBL vsumby3(double [], int);
static LDBL vsumsqr(double [], int);
static LDBL vdotby3(double [], double [], int);


/*=====================================================================*/

static int solve(LDBL e[3],LDBL e0, LDBL det, LDBL spur, LDBL cof) {
  LDBL d, h, g, cth, sth, sqrth;

#define sqrt3 1.732050807568877293527446341506

  d = DSQR(spur);

  h = d - cof;

  if (h > d*epsilon) {
    det *= det;
    g = spur * (cof*1.5 - d) - det*0.5;

    sqrth = DSQRTABS(h);

    d = -g/(h*sqrth);

    if (d > 1-epsilon) {    /* Two identical roots */
      e[0] = spur + 2*sqrth;
      e[2] = e[1] = spur - sqrth;

      return 2;
    } else if (d < -1+epsilon) {    /* Two identical roots */ 
      e[1] = e[0] = spur+sqrth;
      e[2] = MAX(0, spur - 2*sqrth);

      return 3;
    } else {    /* Three distinct roots */
      d = DACOS(d)/3.0;

      cth = sqrth*DCOS(d);
      sth = sqrth*sqrt3*DSIN(d);

      e[0] = spur + 2*cth;
      e[1] = spur - cth + sth;
      e[2] = MAX(0, spur - cth - sth);

      return 1;
    }

  } else {    /* Three identical roots */ 
    e[0] = e[1] = e[2] = spur;

    return 4;
  }
} /* solve */

static void matrix(rotate_t u, xlate_t t, int sw, LDBL e[3],
		   LDBL rr[6], LDBL r[3][3],
		   LDBL sx[3], LDBL sy[3], int npts) {
  LDBL ss[6], d, h, p, a[3][3], b[3][3];
  int i, j, l, m, m1, m2, m3;

  if (sw == 4) {
    /* Case of three identical roots */
    for (i = 0; i < 3; i++) {
      for (j = 0; j < 3; j++) {
        if (i == j)
	  a[i][j] = 1.0;
	else
	  a[i][j] = 0.0;
      }
    }
  } else {
    if (sw == 1) {
      /* Case of three distinct roots */
      for (l=0; l < 2; l++) {
        d = e[l];

        ss[0] = (d-rr[2]) * (d-rr[5]) - DSQR(rr[4]);
        ss[1] = (d-rr[5]) * rr[1] + rr[3]*rr[4];
        ss[2] = (d-rr[0]) * (d-rr[5]) - DSQR(rr[3]);
        ss[3] = (d-rr[2]) * rr[3] + rr[1]*rr[4];
        ss[4] = (d-rr[0]) * rr[4] + rr[1]*rr[3];
        ss[5] = (d-rr[0]) * (d-rr[2]) - DSQR(rr[1]);

        if (FABS(ss[0]) >= FABS(ss[2])) {
          if (FABS(ss[0]) >= FABS(ss[5])) {
            a[0][l] = ss[0];
	    a[1][l] = ss[1];
	    a[2][l] = ss[3];
	  } else {
            a[0][l] = ss[3];
	    a[1][l] = ss[4];
	    a[2][l] = ss[5];
	  }
	} else if (FABS(ss[2]) >= FABS(ss[5])) {
          a[0][l] = ss[1];
	  a[1][l] = ss[2];
	  a[2][l] = ss[4];
	} else {
          a[0][l] = ss[3];
	  a[1][l] = ss[4];
	  a[2][l] = ss[5];
	}

        d = DSQRT(DSQR(a[0][l]) + DSQR(a[1][l]) + DSQR(a[2][l]));

        a[0][l] /= d;
	a[1][l] /= d;
	a[2][l] /= d;
      }

      m1 = 2; m2 = 0; m3 = 1;
    } else {
      /* Cases of two distinct roots */
      if (sw == 2) {
        m = 0;
	m1 = 2;
	m2 = 0;
	m3 = 1;
      } else {
        m = 2;
	m1 = 0;
	m2 = 1;
	m3 = 2;
      }

      h = e[2];

      a[0][1] = 1.0;
      a[1][1] = 1.0;
      a[2][1] = 1.0;

      if (FABS(rr[0]-h) > FABS(rr[2]-h)) {
        if (FABS(rr[0]-h) > FABS(rr[5]-h)) {
          a[0][m] = rr[0]-h;
	  a[1][m] = rr[1];
	  a[2][m] = rr[3];

          p = -(rr[0] + rr[1] + rr[3] - h);

          a[0][1] = p/a[0][m];
	} else {
          a[0][m] = rr[3];
	  a[1][m] = rr[4];
	  a[2][m] = rr[5]-h;

          p = -(rr[3] + rr[4] + rr[5] - h);

          a[2][1] = p/a[2][m];
	}
      } else {
        if (FABS(rr[2]-h) > FABS(rr[5]-h)) {
          a[0][m] = rr[1];
	  a[1][m] = rr[2]-h;
	  a[2][m] = rr[4];

          p = -(rr[1] + rr[2] + rr[4] - h);

          a[1][1] = p/a[1][m];
	} else {
          a[0][m] = rr[3];
	  a[1][m] = rr[4];
	  a[2][m] = rr[5]-h;

          p = -(rr[3] + rr[4] + rr[5] - h);

          a[2][1] = p/a[2][m];
	}
      }

      d = sqrt(DSQR(a[0][m]) + DSQR(a[1][m]) + DSQR(a[2][m]));
      p = sqrt(DSQR(a[0][1]) + DSQR(a[1][1]) + DSQR(a[2][1]));

      for (i = 0; i < 3; i++) {
        a[i][1] /= p;
        a[i][m] /= d;
      }
    }

    /* Common for either two or three distinct roots */

    a[0][m1] = a[1][m2]*a[2][m3] - a[1][m3]*a[2][m2];
    a[1][m1] = a[2][m2]*a[0][m3] - a[2][m3]*a[0][m2];
    a[2][m1] = a[0][m2]*a[1][m3] - a[0][m3]*a[1][m2];
  }

  for (l = 0; l < 2; l++) {
    d = 0.0;

    for (i = 0; i < 3; i++) {
      b[i][l] = r[0][i]*a[0][l] + r[1][i]*a[1][l] + r[2][i]*a[2][l];

      d += SQR(b[i][l]);
    }

    d = DSQRT(d);

    if (d == 0.0) {
      fprintf(stderr,"matrix: degenerate rotation matrix!\n");
      exit(1);
    } else {
      for (i = 0; i < 3; i++)
	b[i][l] /= d;
    }
  }

  b[0][2] = b[1][0]*b[2][1] - b[1][1]*b[2][0];
  b[1][2] = b[2][0]*b[0][1] - b[2][1]*b[0][0];
  b[2][2] = b[0][0]*b[1][1] - b[0][1]*b[1][0];

  /* Calculate rotation matrix */

  for (i = 0; i < 3; i++) {
    for (j = 0; j < 3; j++)
      u[i][j] = (double)(b[i][0]*a[j][0] + b[i][1]*a[j][1] + b[i][2]*a[j][2]);
  }

  /* Calculate translation vector */

  for (i = 0; i < 3; i++)
    t[i] =
      (double)((sy[i] - u[i][0]*sx[0] - u[i][1]*sx[1] - u[i][2]*sx[2])/npts);

  /* return 7; */
}


double bestfit(rotate_t u, xlate_t t, CoordType *a, CoordType *b,
	       const int npts) {
  LDBL sx[3], sy[3], sxy[3][3], sx2, sy2;
  LDBL r[3][3], rr[6];
  LDBL d, spur, det, cof, e0, e[3], inpts;
  int i, j, m, sw;

  (void)vsumby3(&a[0].x, npts);
  (void)vsumby3(&a[0].y, npts);
  (void)vsumby3(&a[0].z, npts);

  (void)vsumby3(&b[0].x, npts);
  (void)vsumby3(&b[0].y, npts);
  (void)vsumby3(&b[0].z, npts);

  (void)vsumsqr(&a[0].x, 3*npts);
  (void)vsumsqr(&b[0].x, 3*npts);

  (void)vdotby3(&a[0].x, &b[0].x, npts);
  (void)vdotby3(&a[0].x, &b[0].y, npts);
  (void)vdotby3(&a[0].x, &b[0].z, npts);
  (void)vdotby3(&a[0].y, &b[0].x, npts);
  (void)vdotby3(&a[0].y, &b[0].y, npts);
  (void)vdotby3(&a[0].y, &b[0].z, npts);
  (void)vdotby3(&a[0].z, &b[0].x, npts);
  (void)vdotby3(&a[0].z, &b[0].y, npts);
  (void)vdotby3(&a[0].z, &b[0].z, npts);

  sx[0] = vsumby3(&a[0].x, npts);
  sx[1] = vsumby3(&a[0].y, npts);
  sx[2] = vsumby3(&a[0].z, npts);
  sy[0] = vsumby3(&b[0].x, npts);
  sy[1] = vsumby3(&b[0].y, npts);
  sy[2] = vsumby3(&b[0].z, npts);
  sx2   = vsumsqr(&a[0].x, 3*npts);
  sy2   = vsumsqr(&b[0].x, 3*npts);
  sxy[0][0] = vdotby3(&a[0].x, &b[0].x, npts);
  sxy[0][1] = vdotby3(&a[0].x, &b[0].y, npts);
  sxy[0][2] = vdotby3(&a[0].x, &b[0].z, npts);
  sxy[1][0] = vdotby3(&a[0].y, &b[0].x, npts);
  sxy[1][1] = vdotby3(&a[0].y, &b[0].y, npts);
  sxy[1][2] = vdotby3(&a[0].y, &b[0].z, npts);
  sxy[2][0] = vdotby3(&a[0].z, &b[0].x, npts);
  sxy[2][1] = vdotby3(&a[0].z, &b[0].y, npts);
  sxy[2][2] = vdotby3(&a[0].z, &b[0].z, npts);

  inpts = 1.0/npts;
  e0 = (sx2 - (DSQR(sx[0])+DSQR(sx[1])+DSQR(sx[2]))*inpts +
        sy2 - (DSQR(sy[0])+DSQR(sy[1])+DSQR(sy[2]))*inpts)*inpts;

  for (i = 0; i < 3; i++) {
    for (j = 0; j < 3; j++) {
      r[i][j] = (sxy[i][j] - sx[i]*sy[j]*inpts)*inpts;
    }
  }
    
  for (m = j = 0; j < 3; j++) {
    for (i = 0; i <= j; i++,m++) {
      rr[m] = r[i][0]*r[j][0] + r[i][1]*r[j][1] + r[i][2]*r[j][2];
    }
  }

  det = r[0][0] * (r[1][1]*r[2][2] - r[2][1]*r[1][2])
      - r[1][0] * (r[0][1]*r[2][2] - r[2][1]*r[0][2])
      + r[2][0] * (r[0][1]*r[1][2] - r[1][1]*r[0][2]);

  spur = (rr[0]+rr[2]+rr[5])/3.0;
  cof = (rr[2]*rr[5] - DSQR(rr[4]) + rr[0]*rr[5] -
         DSQR(rr[3]) + rr[0]*rr[2] - DSQR(rr[1]))/3.0;

  sw = solve(e, e0, det, spur, cof);
  matrix(u, t, sw, e, rr, r, sx, sy, npts);
  d = DSQRTABS(e[2]);
  if (det < 0) d = -d;
  d += DSQRTABS(e[1]) + DSQRTABS(e[0]);
  d = e0-2*d;
  return (double)DSQRTABS(d);
} /* bestfit */

int check_eq(int *iina, int *jinb, int *iina_sav, int *jinb_sav, int nentry) {
  int i;

  for (i=0; i<=nentry; i++) {
    if ((iina[i] != iina_sav[i]) || (jinb[i] != jinb_sav[i]))
      return i+1; /* hack to return what was different - Allen */
  }

  return 0;
} /* check_eq */

double distrms(CoordType *a, CoordType *b, int npts) {
  int i, j;
  LDBL ad, bd, sum;

  sum = 0.0;

  for (i = 0; i < npts; i++) {
    for (j = 0; j < i; j++) {
      ad = DSQRT(SQR(a[i].x-a[j].x) + SQR(a[i].y-a[j].y) +
		 SQR(a[i].z-a[j].z));
      bd = DSQRT(SQR(b[i].x-b[j].x) + SQR(b[i].y-b[j].y) +
		 SQR(b[i].z-b[j].z));

      sum += DSQR(ad-bd);
    }
  }

  return (double)DSQRT(2*sum/(npts*(npts-1)));
} /* distrms */

double findCM(double **x_str, int Npts, double *Tcm_str) {
  int i;
  LDBL SumX=0, SumY=0, SumZ=0;

  for (i=0; i<Npts; i++) {
    SumX+=x_str[1][i];
    SumY+=x_str[2][i];
    SumZ+=x_str[3][i];
  } /* for(i<Npts) */

  Tcm_str[1] = (double)(SumX/Npts);
  Tcm_str[2] = (double)(SumY/Npts);
  Tcm_str[3] = (double)(SumZ/Npts);

  return (double)(SumZ/Npts);
} /* findCM */

double rotrans(double **x_str, int size_str, rotate_t u, xlate_t t) {
  int i;
  double xx, yy, zz;

  for (i=0; i<size_str; i++) {
    xx = x_str[1][i];
    yy = x_str[2][i];
    zz = x_str[3][i]; 

    x_str[1][i] = u[0][0]*xx + u[0][1]*yy + u[0][2]*zz;
    x_str[2][i] = u[1][0]*xx + u[1][1]*yy + u[1][2]*zz;
    x_str[3][i] = u[2][0]*xx + u[2][1]*yy + u[2][2]*zz;

    x_str[1][i] +=  t[0];
    x_str[2][i] +=  t[1];
    x_str[3][i] +=  t[2];
  } /* for (i<size_str) */

  printf("Rotating query: %.10f %.10f %.10f; %.10f %.10f %.10f; %.10f %.10f %.10f\n",
	 u[0][0],u[0][1],u[0][2],
	 u[1][0],u[1][1],u[1][2],
	 u[2][0],u[2][1],u[2][2]);
  printf("Moving query: %.10f %.10f %.10f\n", t[0], t[1], t[2]);

  return 7;
} /* rotrans */

static double TRvec(double *TransVec, double **RotMatr, double *Vec) {
  double x,y,z;

  x = Vec[1];
  y = Vec[2];
  z = Vec[3];

  Vec[1] = RotMatr[1][1]*x + RotMatr[1][2]*y + RotMatr[1][3]*z;
  Vec[2] = RotMatr[2][1]*x + RotMatr[2][2]*y + RotMatr[2][3]*z;
  Vec[3] = RotMatr[3][1]*x + RotMatr[3][2]*y + RotMatr[3][3]*z;

  Vec[1] += TransVec[1];
  Vec[2] += TransVec[2];
  Vec[3] += TransVec[3];

  return Vec[3];
}

double TRmoveXYZstruc(double *TransVec, double **RotMatr,
		      double **x_str, int Npts) {
  int i;
  double ret_val;
  double *tempVec;

  tempVec = dVector(1,3);

  for (i=0; i<Npts; i++) {
    tempVec[1] = x_str[1][i]; 
    tempVec[2] = x_str[2][i]; 
    tempVec[3] = x_str[3][i];

    ret_val = TRvec(TransVec, RotMatr, tempVec);

    x_str[1][i] = tempVec[1];
    x_str[2][i] = tempVec[2];
    x_str[3][i] = tempVec[3];
  }

  free_dVector(tempVec,1,3);

  return ret_val;
} /* TRmoveXYZstruc */

int FindMaxXYZDistIdxLast(double **x_str, int Npts) {
   int i, iMaxDist=0;
   LDBL MaxDist, dist;

   MaxDist = DSQRT(SQR(x_str[1][0]) + SQR(x_str[2][0]) + SQR(x_str[3][0]));

   for (i=1; i<Npts; i++) {
     dist = DSQRT(SQR(x_str[1][i]) + SQR(x_str[2][i]) + SQR(x_str[3][i]));
     if (dist > MaxDist) {
       iMaxDist = i;
       MaxDist = dist;
     }
   }

   return iMaxDist;
} /* FindMaxXYZDistIdxLast */

int FindMaxXYDistIdxLast(double **x_str, int Npts) {
   int i, iMaxDist=0;
   LDBL MaxDist, dist;

   MaxDist = DSQRT(SQR(x_str[1][0]) + SQR(x_str[2][0]) );

   for (i=1; i<Npts; i++) {
     dist  = DSQRT(SQR(x_str[1][i]) + SQR(x_str[2][i]) );
     if (dist > MaxDist) {
       iMaxDist = i;
       MaxDist = dist;
     }
   }

   return iMaxDist;
} /* FindMaxXYDistIdxLast */

double AlignZ(double x, double y, double z, double *angle_b, double *angle_c) {
  LDBL tmp_angle_b;

  tmp_angle_b = DATAN(-x/z);

  *angle_b = (double)tmp_angle_b;

  *angle_c = (double)DATAN(-y/(z*DCOS(tmp_angle_b)-x*DCOS(tmp_angle_b)));

  return (double)DCOS(tmp_angle_b);
}

double FormRotMatrZ(double **Rtemp, double angle_a) {
  Rtemp[1][1] = cos(angle_a);
  Rtemp[1][2] = sin(angle_a);
  Rtemp[1][3] = 0.0e0; 

  Rtemp[2][1] = -Rtemp[1][2]; /* sin(angle_a) */
  Rtemp[2][2] = Rtemp[1][1]; /* cos(angle_a) */
  Rtemp[2][3] = 0.0e0; 

  Rtemp[3][1] = 0.0e0;
  Rtemp[3][2] = 0.0e0;
  Rtemp[3][3] = 1.0e0;

  return Rtemp[1][1]+Rtemp[2][2]+Rtemp[3][3];
} /* FormRotMatrZ */

double FormRotMatrY(double **Rtemp, double angle_a) {
  Rtemp[1][1] = cos(angle_a);
  Rtemp[1][2] = 0.0e0;
  Rtemp[1][3] = sin(angle_a); 

  Rtemp[2][1] = 0.0e0;
  Rtemp[2][2] = 1.0e0;
  Rtemp[2][3] = 0.0e0; 

  Rtemp[3][1] = -Rtemp[1][3]; /* sin(angle_a) */
  Rtemp[3][2] = 0.0e0;
  Rtemp[3][3] = Rtemp[1][1]; /* cos(angle_a) */

  return Rtemp[1][1]+Rtemp[2][2]+Rtemp[3][3];
} /* FormRotMatrY */

double FormRotMatrX(double **Rtemp, double angle_a) {
  Rtemp[1][1] = 1.0e0;
  Rtemp[1][2] = 0.0e0;
  Rtemp[1][3] = 0.0e0; 

  Rtemp[2][1] = 0.0e0;
  Rtemp[2][2] = cos(angle_a);
  Rtemp[2][3] = sin(angle_a); 

  Rtemp[3][1] = 0.0e0;
  Rtemp[3][2] = -Rtemp[2][3]; /* sin(angle_a) */
  Rtemp[3][3] = Rtemp[2][2]; /* cos(angle_a) */

  return Rtemp[1][1]+Rtemp[2][2]+Rtemp[3][3];
} /* FormRotMatrX */

/* double MAX2(double a, double b) { */
/*   if ( a>b ) return a; */
/*   return b; */
/* } */

/* double MIN2(double a, double b) { */
/*   if ( a<b ) return a; */
/*   return b; */
/* } */

/*=====================================================================*/
/* Returns sum(i=0..n-1, v[3*i]) */
static LDBL vsumby3(double v[], int n) {
  LDBL sum;
  int i;

  sum = *v;
  n--;
  v += 3;

  for (i = 0; i < n; i++, v += 3)
    sum += *v;

  return sum;
}

/*=====================================================================*/
/* Returns sum(i=0..n-1, v[i]^2) */
static LDBL vsumsqr(double v[], int n) {
  LDBL sum;
  int i;

  sum = DSQR(*v);
  n--;
  v++;

  for (i = 0; i < n; i++, v++)
    sum += DSQR(*v);

  return sum;
}

/*=====================================================================*/
/* Returns sum(i=0..n-1, a[3*i]*b[3*i]) */
static LDBL vdotby3(double a[], double b[],  int n) {
  LDBL sum;
  int i;

  sum = (LDBL)(*a) * (LDBL)(*b);
  n--;
  a += 3;
  b += 3;

  for (i = 0; i < n; i++, a += 3, b += 3)
    sum += (LDBL)(*a) * (LDBL)(*b);

  return sum;
}

/*=====================================================================*/

/* Translation table from ASTRAL (http://astral.berkeley.edu) - 11/12/04
  ala a val v phe f pro p met m ile i leu l asp d glu e lys k
  arg r ser s thr t tyr y his h cys c asn n gln q trp w gly g
  2as d 3ah h 5hp e acl r aib a alm a alo t aly k arm r asa d
  asb d ask d asl d asq d aya a bcs c bhd d bmt t bnn a buc c
  bug l c5c c c6c c ccs c cea c chg a cle l cme c csd a cso c
  csp c css c csw c cxm m cy1 c cy3 c cyg c cym c cyq c dah f
  dal a dar r das d dcy c dgl e dgn q dha a dhi h dil i div v
  dle l dly k dnp a dpn f dpr p dsn s dsp d dth t dtr w dty y
  dva v efc c fla a fme m ggl e glz g gma e gsc g hac a har r 
  hic h hip h hmr r hpq f htr w hyp p iil i iyr y kcx k llp k
  lly k ltr w lym k lyz k maa a men n mhs h mis s mle l mpq g
  msa g mse m mva v nem h nep h nle l nln l nlp l nmc g oas s
  ocs c omt m paq y pca e pec c phi f phl f pr3 c prr a ptr y
  sac s sar g sch c scs c scy c sel s sep s set s shc c shr k
  soc c sty y sva s tih a tpl w tpo t tpq a trg k tro w tyb y
  tyq y tys y tyy y agm r gl3 g smc c asx b cgu e csx c glx z

  Some also added from .raf edits file (those noted as done automatically).

  pca changed to z due to sometimes being Q (GLN) instead of E (GLU)

*/

char convert321(char *s, int warn, int altered) {
  switch (s[0]) {

  case 'A':
    if (s[1] == 'L') {
      if (s[2] == 'A') return 'A';
      if (s[2] == 'O') return 'T';
      if (s[2] == 'Y') return 'K';
      break;
    }
    if (s[1] == 'R') {
      if (s[2] == 'G') return 'R';
      if (s[2] == 'M') return 'R';
      break;
    }
    if (s[1] == 'S') {
      if (s[2] == 'N') return 'N';
      if (s[2] == 'X') return 'B';
      if ((s[2] == 'P') ||
	  (s[2] == 'A') ||
	  (s[2] == 'B') ||
	  (s[2] == 'K') ||
	  (s[2] == 'L') ||
	  (s[2] == 'Q'))
	return 'D';
      break;
    }
    if (s[1] == 'C') {
      if (s[2] == 'L') return 'R';
      if (s[2] == 'A') return 'A';
      if (s[2] == 'Y') return 'G';
      break;
    }
    if (s[1] == 'G' && s[2] == 'M') return 'R';
    if (s[1] == 'I' && s[2] == 'B') return 'A';
    if (s[1] == 'Y' && s[2] == 'A') return 'A';
    if (s[1] == 'A' && s[2] == 'R') return 'R';
    if (s[1] == 'B' && s[2] == 'A') return 'B';
    break;

  case 'B':
    if (s[1] == 'C' && s[2] == 'S') return 'C';
    if (s[1] == 'H' && s[2] == 'D') return 'D';
    if (s[1] == 'M' && s[2] == 'T') return 'T';
    if (s[1] == 'N' && s[2] == 'N') return 'A';
    if (s[1] == 'T') {
      if (s[2] == 'A') return 'E';
      if (s[2] == 'C') return 'G';
      break;
    }
    if (s[1] == 'U') {
      if (s[2] == 'C') return 'C';
      if (s[2] == 'G') return 'L';
      break;
    }
    break;

  case 'C':
    if (s[1] == 'Y') {
      if ((s[2] == 'S') ||
	  (s[2] == '1') ||
	  (s[2] == '3') ||
	  (s[2] == 'G') ||
	  (s[2] == 'M') ||
	  (s[2] == 'Q'))
	return 'C';
      break;
    }
    if (s[1] == 'S') {
      if ((s[2] == 'B') ||
	  (s[2] == 'E') ||
	  (s[2] == 'O') ||
	  (s[2] == 'P') ||
	  (s[2] == 'S') ||
	  (s[2] == 'W') ||
	  (s[2] == 'X'))
	return 'C';
      if (s[2] == 'D') return 'A';
      break;
    }
    if (s[1] == 'L') {
      if ((s[2] == 'B') ||
	  (s[2] == 'D'))
	return 'S';
      if (s[2] == 'E')
	return 'L';
      break;
    }
    if (s[1] == '5' && s[2] == 'C') return 'C';
    if (s[1] == '6' && s[2] == 'C') return 'C';
    if (s[1] == 'A' && s[2] == 'S') return 'C';
    if (s[1] == 'A' && s[2] == 'Y') return 'C';
    if (s[1] == 'C' && s[2] == 'S') return 'C';
    if (s[1] == 'E' && s[2] == 'A') return 'C';
    if (s[1] == 'G' && s[2] == 'U') return 'E';
    if (s[1] == 'H' && s[2] == 'G') return 'A';
    if (s[1] == 'M' && s[2] == 'E') return 'C';
    if (s[1] == 'X' && s[2] == 'M') return 'M';
    break;

  case 'D':
    if (s[1] == 'A') {
      if (s[2] == 'H') return 'F';
      if (s[2] == 'L') return 'A';
      if (s[2] == 'R') return 'R';
      if (s[2] == 'S') return 'D';
      break;
    }
    if (s[1] == 'T') {
      if (s[2] == 'H') return 'T';
      if (s[2] == 'R') return 'W';
      if (s[2] == 'Y') return 'Y';
      break;
    }
    if (s[1] == 'C' && s[2] == 'Y') return 'C';
    if (s[1] == 'G' && s[2] == 'L') return 'E';
    if (s[1] == 'G' && s[2] == 'N') return 'Q';
    if (s[1] == 'H' && s[2] == 'A') return 'A';
    if (s[1] == 'H' && s[2] == 'I') return 'H';
    if (s[1] == 'I' && s[2] == 'L') return 'I';
    if (s[1] == 'I' && s[2] == 'V') return 'V';
    if (s[1] == 'L' && s[2] == 'E') return 'L';
    if (s[1] == 'L' && s[2] == 'Y') return 'K';
    if (s[1] == 'N' && s[2] == 'P') return 'A';
    if (s[1] == 'O' && s[2] == 'H') return 'D';
    if (s[1] == 'P' && s[2] == 'N') return 'F';
    if (s[1] == 'P' && s[2] == 'R') return 'P';
    if (s[1] == 'S' && s[2] == 'N') return 'S';
    if (s[1] == 'S' && s[2] == 'P') return 'D';
    if (s[1] == 'V' && s[2] == 'A') return 'V';
    break;

  case 'F':
    if (s[1] == 'G' && s[2] == 'L') return 'C';
    if (s[1] == 'L' && s[2] == 'A') return 'A';
    if (s[1] == 'M' && s[2] == 'E') return 'M';
    break;

  case 'G':
    if (s[1] == 'L') {
      if (s[2] == 'N') return 'Q';    
      if (s[2] == 'U') return 'E';
      if (s[2] == 'Y') return 'G';
      if (s[2] == 'X') return 'Z';
      if ((s[2] == 'Y') ||
	  (s[2] == '3') ||
	  (s[2] == 'Z'))
	return 'G';
      break;
    }
    if (s[1] == 'G' && s[2] == 'L') return 'E';
    if (s[1] == 'M' && s[2] == 'A') return 'E';
    if (s[1] == 'S' && s[2] == 'C') return 'G';
    break;

  case 'H':
    if (s[1] == 'I') {
      if ((s[2] == 'S') ||
	  (s[2] == 'C') ||
	  (s[2] == 'P'))
	return 'H';
      break;
    }
    if (s[1] == 'A' && s[2] == 'C') return 'A';
    if (s[1] == 'A' && s[2] == 'R') return 'R';
    if (s[1] == 'M' && s[2] == 'R') return 'R';
    if (s[1] == 'P' && s[2] == 'Q') return 'F';
    if (s[1] == 'T' && s[2] == 'R') return 'W';
    if (s[1] == 'Y' && s[2] == 'P') return 'P';
    break;

  case 'I':
    if (s[1] == 'A' && s[2] == 'S') return 'B';
    if (s[1] == 'I' && s[2] == 'L') return 'I';
    if (s[1] == 'N' && s[2] == 'I') return 'K';
    if (s[1] == 'Y' && s[2] == 'R') return 'Y';
    if (s[1] == 'L' && s[2] == 'E') return 'I';
    break;

  case 'L':
    if (s[1] == 'Y') {
      if ((s[2] == 'S') ||
	  (s[2] == 'M') ||
	  (s[2] == 'X') ||
	  (s[2] == 'Z'))
	return 'K';
      break;
    }
    if (s[1] == 'E' && s[2] == 'U') return 'L';    
    if (s[1] == 'L') {
      if ((s[2] == 'P') ||
	  (s[2] == 'Y'))
	return 'K';
      break;
    }
    if (s[1] == 'T' && s[2] == 'R') return 'W';
    break;

  case 'M':
    if (s[1] == 'E' && s[2] == 'T') return 'M';
    if (s[1] == 'L') {
      if ((s[2] == 'Y') ||
	  (s[2] == 'Z'))
	return 'K';
      if (s[2] == 'E') return 'L';
      break;
    }
    if (s[1] == '3' && s[2] == 'L') return 'K';
    if (s[1] == 'A' && s[2] == 'A') return 'A';
    if (s[1] == 'E' && s[2] == 'N') return 'N';
    if (s[1] == 'H' && s[2] == 'O') return 'M';
    if (s[1] == 'H' && s[2] == 'S') return 'H';
    if (s[1] == 'I' && s[2] == 'S') return 'S';
    if (s[1] == 'P' && s[2] == 'Q') return 'G';
    if (s[1] == 'S' && s[2] == 'A') return 'G';
    if (s[1] == 'S' && s[2] == 'E') return 'M';
    if (s[1] == 'V' && s[2] == 'A') return 'V';
    break;

  case 'N':
    if (s[1] == 'E') {
      if ((s[2] == 'M') ||
	  (s[2] == 'P'))
	return 'H';
      break;
    }
    if (s[1] == 'L') {
      if ((s[2] == 'E') ||
	  (s[2] == 'N') ||
	  (s[2] == 'P'))
	return 'L';
      break;
    }
    if (s[1] == 'M' && s[2] == 'C') return 'G';
    if (s[1] == 'P' && s[2] == 'H') return 'C';
    break;

  case 'O':
    if (s[1] == 'A' && s[2] == 'S') return 'S';
    if (s[1] == 'C') {
      if ((s[2] == 'S') ||
	  (s[2] == 'Y'))
	return 'C';
      break;
    }
    if (s[1] == 'M' && s[2] == 'T') return 'M';
    break;

  case 'P':
    if (s[1] == 'H') {
      if ((s[2] == 'E') ||
	  (s[2] == 'I') ||
	  (s[2] == 'L'))
	return 'F';
      break;
    }
    if (s[1] == 'R') {
      if (s[2] == 'O') return 'P';
      if (s[2] == 'S') return 'P';
      if (s[2] == '3') return 'C';
      if (s[2] == 'R') return 'A';
      break;
    }
    if (s[1] == 'A' && s[2] == 'Q') return 'Y';
    if (s[1] == 'C' && s[2] == 'A') return 'Z';
    if (s[1] == 'E' && s[2] == 'C') return 'C';
    if (s[1] == 'T' && s[2] == 'R') return 'Y';
    if (s[1] == 'Y' && s[2] == 'X') return 'C';
    break;

  case 'S':
    if (s[1] == 'E') {
      if ((s[2] == 'R') ||
	  (s[2] == 'L') ||
	  (s[2] == 'P') ||
	  (s[2] == 'T'))
	return 'S';
      break;
    }
    if (s[1] == 'C') {
      if ((s[2] == 'H') ||
	  (s[2] == 'S') ||
	  (s[2] == 'Y'))
	return 'C';
      break;
    }
    if (s[1] == 'B') {
      if ((s[2] == 'D') ||
	  (s[2] == 'L'))
	return 'S';
      break;
    }
    if (s[1] == 'A' && s[2] == 'C') return 'S';
    if (s[1] == 'A' && s[2] == 'R') return 'G';
    if (s[1] == 'H' && s[2] == 'C') return 'C';
    if (s[1] == 'H' && s[2] == 'R') return 'K';
    if (s[1] == 'M' && s[2] == 'C') return 'C';
    if (s[1] == 'M' && s[2] == 'E') return 'M';
    if (s[1] == 'N' && s[2] == 'N') return 'D';
    if (s[1] == 'O' && s[2] == 'C') return 'C';
    if (s[1] == 'T' && s[2] == 'Y') return 'Y';
    if (s[1] == 'V' && s[2] == 'A') return 'S';
    break;

  case 'T':
    if (s[1] == 'Y') {
      if ((s[2] == 'R') ||
	  (s[2] == 'B') ||
	  (s[2] == 'Q') ||
	  (s[2] == 'S') ||
	  (s[2] == 'I') ||
	  (s[2] == 'Y'))
	return 'Y';
      break;
    }
    if (s[1] == 'R') {
      if ((s[2] == 'P') ||
	  (s[2] == 'N') ||
	  (s[2] == 'O'))
	return 'W';
      if (s[2] == 'G') return 'K';
      break;
    }
    if (s[1] == 'H' && s[2] == 'R') return 'T';   
    if (s[1] == 'P') {
      if (s[2] == 'L') return 'W';
      if (s[2] == 'O') return 'T';
      if (s[2] == 'Q') return 'A';
      break;
    }
    if (s[1] == 'I' && s[2] == 'H') return 'A';
    break;

  default:
    if (s[0] == 'V' && s[1] == 'A' && s[2] == 'L') return 'V';
    if (s[0] == '2' && s[1] == 'A' && s[2] == 'S') return 'D';
    if (s[0] == '3' && s[1] == 'A' && s[2] == 'H') return 'H';
    if (s[0] == '5' && s[1] == 'H' && s[2] == 'P') return 'E';
    if (s[0] == 'E' && s[1] == 'F' && s[2] == 'C') return 'C';
    if (s[0] == 'K' && s[1] == 'C' && s[2] == 'X') return 'K';
    if (s[0] == 'U' && s[1] == 'N' && s[2] == 'K') return 'X';
    if (altered && s[0] == '_' && s[1] == '_' && (! isalpha(s[2])))
      return s[2];
    break;
  }

  if (warn) {
    fprintf(stderr, "convert321: conversion of %s did not succeed!\n", s);
    return 'X';
  } else
    return '?';
} /* convert321 */

/* from "Amino Acid Exchange Matrices" article, page 826 */

double gonnet_p(char a, char b, int altered, int use_carbonyl) {
  if (use_carbonyl) {
    if (isascii(a) != isascii(b))
      return -1000;
    else {
      if (! isascii(a)) {
	a = toascii(a);
	b = toascii(b);
	if (isalpha(a) != isalpha(b))
	  return -1000;
      }
      if (isalpha(a) && isalpha(b)) {
	if (isupper(a) != isupper(b))
	  return -1000;
	else if (islower(a)) {
	  a = _toupper(a);
	  b = _toupper(b);
	}
      }
    }
  }

  if (altered && (! (isalpha(a) && isalpha(b)))) {
    if (a == b)
      return 1000;
    else
      return -1000;
  }

  if (a < b) return gonnet_p(b,a,0,0);

  if ((a == 'A') && (b == 'A')) return 7.6;
  
  if ((a == 'B') && (b == 'A')) return 4.9;
  if ((a == 'B') && (b == 'B')) return 9.9;
  
  if ((a == 'C') && (b == 'A')) return 5.7;
  if ((a == 'C') && (b == 'B')) return 2.7;
  if ((a == 'C') && (b == 'C')) return 16.7;
  
  if ((a == 'D') && (b == 'A')) return 4.9;
  if ((a == 'D') && (b == 'B')) return 8.65;
  if ((a == 'D') && (b == 'C')) return 2;
  if ((a == 'D') && (b == 'D')) return 9;

  if ((a == 'E') && (b == 'A')) return 5.2;
  if ((a == 'E') && (b == 'B')) return 7;
  if ((a == 'E') && (b == 'C')) return 2.2;
  if ((a == 'E') && (b == 'D')) return 7.9;
  if ((a == 'E') && (b == 'E')) return 8.8;

  if ((a == 'F') && (b == 'A')) return 2.9;
  if ((a == 'F') && (b == 'B')) return 1.4;
  if ((a == 'F') && (b == 'C')) return 4.4;
  if ((a == 'F') && (b == 'D')) return 0.7;
  if ((a == 'F') && (b == 'E')) return 1.3;
  if ((a == 'F') && (b == 'F')) return 12.2;

  if ((a == 'G') && (b == 'A')) return 5.7;
  if ((a == 'G') && (b == 'B')) return 5.45;
  if ((a == 'G') && (b == 'C')) return 3.2;
  if ((a == 'G') && (b == 'D')) return 5.3;
  if ((a == 'G') && (b == 'E')) return 4.4;
  if ((a == 'G') && (b == 'F')) return 0;
  if ((a == 'G') && (b == 'G')) return 11.8;

  if ((a == 'H') && (b == 'A')) return 4.4;
  if ((a == 'H') && (b == 'B')) return 6;
  if ((a == 'H') && (b == 'C')) return 3.9;
  if ((a == 'H') && (b == 'D')) return 5.6;
  if ((a == 'H') && (b == 'E')) return 5.6;
  if ((a == 'H') && (b == 'F')) return 5.1;
  if ((a == 'H') && (b == 'G')) return 3.8;
  if ((a == 'H') && (b == 'H')) return 11.2;

  if ((a == 'I') && (b == 'A')) return 4.4;
  if ((a == 'I') && (b == 'B')) return 1.9;
  if ((a == 'I') && (b == 'C')) return 4.1;
  if ((a == 'I') && (b == 'D')) return 1.4;
  if ((a == 'I') && (b == 'E')) return 2.5;
  if ((a == 'I') && (b == 'F')) return 6.2;
  if ((a == 'I') && (b == 'G')) return 0.7;
  if ((a == 'I') && (b == 'H')) return 3;
  if ((a == 'I') && (b == 'I')) return 9.2;

  if ((a == 'K') && (b == 'A')) return 4.8;
  if ((a == 'K') && (b == 'B')) return 5.85;
  if ((a == 'K') && (b == 'C')) return 2.4;
  if ((a == 'K') && (b == 'D')) return 5.7;
  if ((a == 'K') && (b == 'E')) return 6.4;
  if ((a == 'K') && (b == 'F')) return 1.9;
  if ((a == 'K') && (b == 'G')) return 4.1;
  if ((a == 'K') && (b == 'H')) return 5.8;
  if ((a == 'K') && (b == 'I')) return 3.1;
  if ((a == 'K') && (b == 'K')) return 8.4;

  if ((a == 'L') && (b == 'A')) return 4;
  if ((a == 'L') && (b == 'B')) return 1.7;
  if ((a == 'L') && (b == 'C')) return 3.7;
  if ((a == 'L') && (b == 'D')) return 1.2;
  if ((a == 'L') && (b == 'E')) return 2.4;
  if ((a == 'L') && (b == 'F')) return 7.2;
  if ((a == 'L') && (b == 'G')) return 0.8;
  if ((a == 'L') && (b == 'H')) return 3.3;
  if ((a == 'L') && (b == 'I')) return 8;
  if ((a == 'L') && (b == 'K')) return 3.1;
  if ((a == 'L') && (b == 'L')) return 9.2;

  if ((a == 'M') && (b == 'A')) return 4.5;
  if ((a == 'M') && (b == 'B')) return 2.6;
  if ((a == 'M') && (b == 'C')) return 4.3;
  if ((a == 'M') && (b == 'D')) return 2.2;
  if ((a == 'M') && (b == 'E')) return 3.2;
  if ((a == 'M') && (b == 'F')) return 6.8;
  if ((a == 'M') && (b == 'G')) return 1.7;
  if ((a == 'M') && (b == 'H')) return 3.9;
  if ((a == 'M') && (b == 'I')) return 7.7;
  if ((a == 'M') && (b == 'K')) return 3.8;
  if ((a == 'M') && (b == 'L')) return 8.8;
  if ((a == 'M') && (b == 'M')) return 9.5;

  if ((a == 'N') && (b == 'A')) return 4.9;
  if ((a == 'N') && (b == 'B')) return 8.2;
  if ((a == 'N') && (b == 'C')) return 3.4;
  if ((a == 'N') && (b == 'D')) return 7.4;
  if ((a == 'N') && (b == 'E')) return 6.1;
  if ((a == 'N') && (b == 'F')) return 2.1;
  if ((a == 'N') && (b == 'G')) return 5.6;
  if ((a == 'N') && (b == 'H')) return 6.4;
  if ((a == 'N') && (b == 'I')) return 2.4;
  if ((a == 'N') && (b == 'K')) return 6;
  if ((a == 'N') && (b == 'L')) return 2.2;
  if ((a == 'N') && (b == 'M')) return 3;
  if ((a == 'N') && (b == 'N')) return 9;

  if ((a == 'P') && (b == 'A')) return 5.5;
  if ((a == 'P') && (b == 'B')) return 4.4;
  if ((a == 'P') && (b == 'C')) return 2.1;
  if ((a == 'P') && (b == 'D')) return 4.5;
  if ((a == 'P') && (b == 'E')) return 4.7;
  if ((a == 'P') && (b == 'F')) return 1.4;
  if ((a == 'P') && (b == 'G')) return 3.6;
  if ((a == 'P') && (b == 'H')) return 4.1;
  if ((a == 'P') && (b == 'I')) return 2.6;
  if ((a == 'P') && (b == 'K')) return 4.6;
  if ((a == 'P') && (b == 'L')) return 2.9;
  if ((a == 'P') && (b == 'M')) return 2.8;
  if ((a == 'P') && (b == 'N')) return 4.3;
  if ((a == 'P') && (b == 'P')) return 12.8;

  if ((a == 'Q') && (b == 'A')) return 5;
  if ((a == 'Q') && (b == 'B')) return 6;
  if ((a == 'Q') && (b == 'C')) return 2.8;
  if ((a == 'Q') && (b == 'D')) return 6.1;
  if ((a == 'Q') && (b == 'E')) return 6.9;
  if ((a == 'Q') && (b == 'F')) return 2.6;
  if ((a == 'Q') && (b == 'G')) return 4.2;
  if ((a == 'Q') && (b == 'H')) return 6.4;
  if ((a == 'Q') && (b == 'I')) return 3.3;
  if ((a == 'Q') && (b == 'K')) return 6.7;
  if ((a == 'Q') && (b == 'L')) return 3.6;
  if ((a == 'Q') && (b == 'M')) return 4.2;
  if ((a == 'Q') && (b == 'N')) return 5.9;
  if ((a == 'Q') && (b == 'P')) return 5;
  if ((a == 'Q') && (b == 'Q')) return 7.9;

  if ((a == 'R') && (b == 'A')) return 4.6;
  if ((a == 'R') && (b == 'B')) return 5.2;
  if ((a == 'R') && (b == 'C')) return 3;
  if ((a == 'R') && (b == 'D')) return 4.9;
  if ((a == 'R') && (b == 'E')) return 5.6;
  if ((a == 'R') && (b == 'F')) return 2;
  if ((a == 'R') && (b == 'G')) return 4.2;
  if ((a == 'R') && (b == 'H')) return 5.8;
  if ((a == 'R') && (b == 'I')) return 2.8;
  if ((a == 'R') && (b == 'K')) return 7.9;
  if ((a == 'R') && (b == 'L')) return 3;
  if ((a == 'R') && (b == 'M')) return 3.5;
  if ((a == 'R') && (b == 'N')) return 5.5;
  if ((a == 'R') && (b == 'P')) return 4.3;
  if ((a == 'R') && (b == 'Q')) return 6.7;
  if ((a == 'R') && (b == 'R')) return 9.9;

  if ((a == 'S') && (b == 'A')) return 6.3;
  if ((a == 'S') && (b == 'B')) return 5.9;
  if ((a == 'S') && (b == 'C')) return 5.3;
  if ((a == 'S') && (b == 'D')) return 5.7;
  if ((a == 'S') && (b == 'E')) return 5.4;
  if ((a == 'S') && (b == 'F')) return 2.4;
  if ((a == 'S') && (b == 'G')) return 5.6;
  if ((a == 'S') && (b == 'H')) return 5;
  if ((a == 'S') && (b == 'I')) return 3.4;
  if ((a == 'S') && (b == 'K')) return 5.3;
  if ((a == 'S') && (b == 'L')) return 3.1;
  if ((a == 'S') && (b == 'M')) return 3.8;
  if ((a == 'S') && (b == 'N')) return 6.1;
  if ((a == 'S') && (b == 'P')) return 5.6;
  if ((a == 'S') && (b == 'Q')) return 5.4;
  if ((a == 'S') && (b == 'R')) return 5;
  if ((a == 'S') && (b == 'S')) return 7.4;

  if ((a == 'T') && (b == 'A')) return 5.8;
  if ((a == 'T') && (b == 'B')) return 5.45;
  if ((a == 'T') && (b == 'C')) return 4.7;
  if ((a == 'T') && (b == 'D')) return 5.2;
  if ((a == 'T') && (b == 'E')) return 5.1;
  if ((a == 'T') && (b == 'F')) return 3;
  if ((a == 'T') && (b == 'G')) return 4.1;
  if ((a == 'T') && (b == 'H')) return 4.9;
  if ((a == 'T') && (b == 'I')) return 4.6;
  if ((a == 'T') && (b == 'K')) return 5.3;
  if ((a == 'T') && (b == 'L')) return 3.9;
  if ((a == 'T') && (b == 'M')) return 4.6;
  if ((a == 'T') && (b == 'N')) return 5.7;
  if ((a == 'T') && (b == 'P')) return 5.3;
  if ((a == 'T') && (b == 'Q')) return 5.2;
  if ((a == 'T') && (b == 'R')) return 5;
  if ((a == 'T') && (b == 'S')) return 6.7;
  if ((a == 'T') && (b == 'T')) return 7.7;

  if ((a == 'V') && (b == 'A')) return 5.3;
  if ((a == 'V') && (b == 'B')) return 2.65;
  if ((a == 'V') && (b == 'C')) return 5.2;
  if ((a == 'V') && (b == 'D')) return 2.3;
  if ((a == 'V') && (b == 'E')) return 3.3;
  if ((a == 'V') && (b == 'F')) return 5.3;
  if ((a == 'V') && (b == 'G')) return 1.9;
  if ((a == 'V') && (b == 'H')) return 3.2;
  if ((a == 'V') && (b == 'I')) return 8.3;
  if ((a == 'V') && (b == 'K')) return 3.5;
  if ((a == 'V') && (b == 'L')) return 7;
  if ((a == 'V') && (b == 'M')) return 6.8;
  if ((a == 'V') && (b == 'N')) return 3;
  if ((a == 'V') && (b == 'P')) return 3.4;
  if ((a == 'V') && (b == 'Q')) return 3.7;
  if ((a == 'V') && (b == 'R')) return 3.2;
  if ((a == 'V') && (b == 'S')) return 4.2;
  if ((a == 'V') && (b == 'T')) return 5.2;
  if ((a == 'V') && (b == 'V')) return 8.6;

  if ((a == 'W') && (b == 'A')) return 1.6;
  if ((a == 'W') && (b == 'B')) return 0.8;
  if ((a == 'W') && (b == 'C')) return 4.2;
  if ((a == 'W') && (b == 'D')) return 0;
  if ((a == 'W') && (b == 'E')) return 0.9;
  if ((a == 'W') && (b == 'F')) return 8.8;
  if ((a == 'W') && (b == 'G')) return 1.2;
  if ((a == 'W') && (b == 'H')) return 4.4;
  if ((a == 'W') && (b == 'I')) return 3.4;
  if ((a == 'W') && (b == 'K')) return 1.7;
  if ((a == 'W') && (b == 'L')) return 4.5;
  if ((a == 'W') && (b == 'M')) return 4.2;
  if ((a == 'W') && (b == 'N')) return 1.6;
  if ((a == 'W') && (b == 'P')) return 0.2;
  if ((a == 'W') && (b == 'Q')) return 2.5;
  if ((a == 'W') && (b == 'R')) return 3.6;
  if ((a == 'W') && (b == 'S')) return 1.9;
  if ((a == 'W') && (b == 'T')) return 1.7;
  if ((a == 'W') && (b == 'V')) return 2.6;
  if ((a == 'W') && (b == 'W')) return 19.4;

  if ((a == 'Y') && (b == 'A')) return 3;
  if ((a == 'Y') && (b == 'B')) return 3.1;
  if ((a == 'Y') && (b == 'C')) return 4.7;
  if ((a == 'Y') && (b == 'D')) return 2.4;
  if ((a == 'Y') && (b == 'E')) return 2.5;
  if ((a == 'Y') && (b == 'F')) return 10.3;
  if ((a == 'Y') && (b == 'G')) return 1.2;
  if ((a == 'Y') && (b == 'H')) return 7.4;
  if ((a == 'Y') && (b == 'I')) return 4.5;
  if ((a == 'Y') && (b == 'K')) return 3.1;
  if ((a == 'Y') && (b == 'L')) return 5.2;
  if ((a == 'Y') && (b == 'M')) return 5;
  if ((a == 'Y') && (b == 'N')) return 3.8;
  if ((a == 'Y') && (b == 'P')) return 2.1;
  if ((a == 'Y') && (b == 'Q')) return 3.5;
  if ((a == 'Y') && (b == 'R')) return 3.4;
  if ((a == 'Y') && (b == 'S')) return 3.3;
  if ((a == 'Y') && (b == 'T')) return 3.3;
  if ((a == 'Y') && (b == 'V')) return 4.1;
  if ((a == 'Y') && (b == 'W')) return 9.3;
  if ((a == 'Y') && (b == 'Y')) return 13;

  if ((a == 'Z') && (b == 'A')) return 5.1;
  if ((a == 'Z') && (b == 'C')) return 2.5;
  if ((a == 'Z') && (b == 'D')) return 7;
  if ((a == 'Z') && (b == 'E')) return 7.85;
  if ((a == 'Z') && (b == 'F')) return 1.95;
  if ((a == 'Z') && (b == 'G')) return 4.3;
  if ((a == 'Z') && (b == 'H')) return 6;
  if ((a == 'Z') && (b == 'I')) return 2.9;
  if ((a == 'Z') && (b == 'K')) return 6.55;
  if ((a == 'Z') && (b == 'L')) return 3;
  if ((a == 'Z') && (b == 'M')) return 3.7;
  if ((a == 'Z') && (b == 'N')) return 6;
  if ((a == 'Z') && (b == 'P')) return 4.85;
  if ((a == 'Z') && (b == 'Q')) return 7.4;
  if ((a == 'Z') && (b == 'R')) return 6.15;
  if ((a == 'Z') && (b == 'S')) return 5.4;
  if ((a == 'Z') && (b == 'T')) return 5.15;
  if ((a == 'Z') && (b == 'V')) return 3.5;
  if ((a == 'Z') && (b == 'W')) return 1.7;
  if ((a == 'Z') && (b == 'Y')) return 3;
  if ((a == 'Z') && (b == 'Z')) return 7.9;

  if (a == b)
    return 12;
  else
    return 0;
} /* gonnet_p */

double pam120_p(char a, char b, int altered, int use_carbonyl) {
  if (use_carbonyl) {
    if (isascii(a) != isascii(b))
      return -1000;
    else {
      if (! isascii(a)) {
	a = toascii(a);
	b = toascii(b);
	if (isalpha(a) != isalpha(b))
	  return -1000;
      }
      if (isalpha(a) && isalpha(b)) {
	if (isupper(a) != isupper(b))
	  return -1000;
	else if (islower(a)) {
	  a = _toupper(a);
	  b = _toupper(b);
	}
      }
    }
  }

  if (altered && (! (isalpha(a) && isalpha(b)))) {
    if (a == b)
      return 1000;
    else
      return -1000;
  }

  if (a < b) return pam120_p(b,a,0,0);

  if ((a == 'A') && (b == 'A')) return 11.0;
  
  if ((a == 'B') && (b == 'A')) return 8.0;
  if ((a == 'B') && (b == 'B')) return 12.0;
  
  if ((a == 'C') && (b == 'A')) return 5.0;
  if ((a == 'C') && (b == 'B')) return 2.0;
  if ((a == 'C') && (b == 'C')) return 17.0;
  
  if ((a == 'D') && (b == 'A')) return 8.0;
  if ((a == 'D') && (b == 'B')) return 12.0;
  if ((a == 'D') && (b == 'C')) return 1.0;
  if ((a == 'D') && (b == 'D')) return 13.0;
  
  if ((a == 'E') && (b == 'A')) return 8.0;
  if ((a == 'E') && (b == 'B')) return 11.0;
  if ((a == 'E') && (b == 'C')) return 1.0;
  if ((a == 'E') && (b == 'D')) return 11.0;
  if ((a == 'E') && (b == 'E')) return 13.0;
  
  if ((a == 'F') && (b == 'A')) return 4.0;
  if ((a == 'F') && (b == 'B')) return 3.0;
  if ((a == 'F') && (b == 'C')) return 2.0;
  if ((a == 'F') && (b == 'D')) return 1.0;
  if ((a == 'F') && (b == 'E')) return 1.0;
  if ((a == 'F') && (b == 'F')) return 16.0;
  
  if ((a == 'G') && (b == 'A')) return 9.0;
  if ((a == 'G') && (b == 'B')) return 8.0;
  if ((a == 'G') && (b == 'C')) return 4.0;
  if ((a == 'G') && (b == 'D')) return 8.0;
  if ((a == 'G') && (b == 'E')) return 7.0;
  if ((a == 'G') && (b == 'F')) return 3.0;
  if ((a == 'G') && (b == 'G')) return 13.0;
  
  if ((a == 'H') && (b == 'A')) return 5.0;
  if ((a == 'H') && (b == 'B')) return 9.0;
  if ((a == 'H') && (b == 'C')) return 4.0;
  if ((a == 'H') && (b == 'D')) return 8.0;
  if ((a == 'H') && (b == 'E')) return 7.0;
  if ((a == 'H') && (b == 'F')) return 5.0;
  if ((a == 'H') && (b == 'G')) return 4.0;
  if ((a == 'H') && (b == 'H')) return 15.0;
  
  if ((a == 'I') && (b == 'A')) return 7.0;
  if ((a == 'I') && (b == 'B')) return 5.0;
  if ((a == 'I') && (b == 'C')) return 5.0;
  if ((a == 'I') && (b == 'D')) return 5.0;
  if ((a == 'I') && (b == 'E')) return 5.0;
  if ((a == 'I') && (b == 'F')) return 8.0;
  if ((a == 'I') && (b == 'G')) return 4.0;
  if ((a == 'I') && (b == 'H')) return 4.0;
  if ((a == 'I') && (b == 'I')) return 14.0;
  
  if ((a == 'K') && (b == 'A')) return 6.0;
  if ((a == 'K') && (b == 'B')) return 8.0;
  if ((a == 'K') && (b == 'C')) return 1.0;
  if ((a == 'K') && (b == 'D')) return 7.0;
  if ((a == 'K') && (b == 'E')) return 7.0;
  if ((a == 'K') && (b == 'F')) return 1.0;
  if ((a == 'K') && (b == 'G')) return 5.0;
  if ((a == 'K') && (b == 'H')) return 6.0;
  if ((a == 'K') && (b == 'I')) return 5.0;
  if ((a == 'K') && (b == 'K')) return 13.0;
 
  if ((a == 'L') && (b == 'A')) return 5.0;
  if ((a == 'L') && (b == 'B')) return 4.0;
  if ((a == 'L') && (b == 'C')) return 1.0;
  if ((a == 'L') && (b == 'D')) return 3.0;
  if ((a == 'L') && (b == 'E')) return 4.0;
  if ((a == 'L') && (b == 'F')) return 8.0;
  if ((a == 'L') && (b == 'G')) return 3.0;
  if ((a == 'L') && (b == 'H')) return 5.0;
  if ((a == 'L') && (b == 'I')) return 9.0;
  if ((a == 'L') && (b == 'K')) return 4.0;
  if ((a == 'L') && (b == 'L')) return 13.0;

  if ((a == 'M') && (b == 'A')) return 6.0;
  if ((a == 'M') && (b == 'B')) return 4.0;
  if ((a == 'M') && (b == 'C')) return 2.0;
  if ((a == 'M') && (b == 'D')) return 4.0;
  if ((a == 'M') && (b == 'E')) return 5.0;
  if ((a == 'M') && (b == 'F')) return 7.0;
  if ((a == 'M') && (b == 'G')) return 4.0;
  if ((a == 'M') && (b == 'H')) return 4.0;
  if ((a == 'M') && (b == 'I')) return 9.0;
  if ((a == 'M') && (b == 'K')) return 8.0;
  if ((a == 'M') && (b == 'L')) return 11.0;
  if ((a == 'M') && (b == 'M')) return 16.0;

  if ((a == 'N') && (b == 'A')) return 7.0;
  if ((a == 'N') && (b == 'B')) return 11.0;
  if ((a == 'N') && (b == 'C')) return 3.0;
  if ((a == 'N') && (b == 'D')) return 10.0;
  if ((a == 'N') && (b == 'E')) return 9.0;
  if ((a == 'N') && (b == 'F')) return 4.0;
  if ((a == 'N') && (b == 'G')) return 8.0;
  if ((a == 'N') && (b == 'H')) return 10.0;
  if ((a == 'N') && (b == 'I')) return 6.0;
  if ((a == 'N') && (b == 'K')) return 9.0;
  if ((a == 'N') && (b == 'L')) return 4.0;
  if ((a == 'N') && (b == 'M')) return 5.0;
  if ((a == 'N') && (b == 'N')) return 12.0;

  if ((a == 'P') && (b == 'A')) return 9.0;
  if ((a == 'P') && (b == 'B')) return 6.0;
  if ((a == 'P') && (b == 'C')) return 4.0;
  if ((a == 'P') && (b == 'D')) return 5.0;
  if ((a == 'P') && (b == 'E')) return 6.0;
  if ((a == 'P') && (b == 'F')) return 3.0;
  if ((a == 'P') && (b == 'G')) return 6.0;
  if ((a == 'P') && (b == 'H')) return 7.0;
  if ((a == 'P') && (b == 'I')) return 5.0;
  if ((a == 'P') && (b == 'K')) return 6.0;
  if ((a == 'P') && (b == 'L')) return 5.0;
  if ((a == 'P') && (b == 'M')) return 5.0;
  if ((a == 'P') && (b == 'N')) return 6.0;
  if ((a == 'P') && (b == 'P')) return 14.0;

  if ((a == 'Q') && (b == 'A')) return 7.0;
  if ((a == 'Q') && (b == 'B')) return 8.0;
  if ((a == 'Q') && (b == 'C')) return 1.0;
  if ((a == 'Q') && (b == 'D')) return 9.0;
  if ((a == 'Q') && (b == 'E')) return 10.0;
  if ((a == 'Q') && (b == 'F')) return 2.0;
  if ((a == 'Q') && (b == 'G')) return 5.0;
  if ((a == 'Q') && (b == 'H')) return 11.0;
  if ((a == 'Q') && (b == 'I')) return 5.0;
  if ((a == 'Q') && (b == 'K')) return 8.0;
  if ((a == 'Q') && (b == 'L')) return 6.0;
  if ((a == 'Q') && (b == 'M')) return 7.0;
  if ((a == 'Q') && (b == 'N')) return 8.0;
  if ((a == 'Q') && (b == 'P')) return 8.0;
  if ((a == 'Q') && (b == 'Q')) return 14.0;

  if ((a == 'R') && (b == 'A')) return 5.0;
  if ((a == 'R') && (b == 'B')) return 6.0;
  if ((a == 'R') && (b == 'C')) return 4.0;
  if ((a == 'R') && (b == 'D')) return 5.0;
  if ((a == 'R') && (b == 'E')) return 5.0;
  if ((a == 'R') && (b == 'F')) return 3.0;
  if ((a == 'R') && (b == 'G')) return 4.0;
  if ((a == 'R') && (b == 'H')) return 9.0;
  if ((a == 'R') && (b == 'I')) return 6.0;
  if ((a == 'R') && (b == 'K')) return 10.0;
  if ((a == 'R') && (b == 'L')) return 4.0;
  if ((a == 'R') && (b == 'M')) return 7.0;
  if ((a == 'R') && (b == 'N')) return 7.0;
  if ((a == 'R') && (b == 'P')) return 7.0;
  if ((a == 'R') && (b == 'Q')) return 9.0;
  if ((a == 'R') && (b == 'R')) return 14.0;

  if ((a == 'S') && (b == 'A')) return 9.0;
  if ((a == 'S') && (b == 'B')) return 8.0;
  if ((a == 'S') && (b == 'C')) return 8.0;
  if ((a == 'S') && (b == 'D')) return 8.0;
  if ((a == 'S') && (b == 'E')) return 7.0;
  if ((a == 'S') && (b == 'F')) return 5.0;
  if ((a == 'S') && (b == 'G')) return 9.0;
  if ((a == 'S') && (b == 'H')) return 6.0;
  if ((a == 'S') && (b == 'I')) return 6.0;
  if ((a == 'S') && (b == 'K')) return 7.0;
  if ((a == 'S') && (b == 'L')) return 4.0;
  if ((a == 'S') && (b == 'M')) return 6.0;
  if ((a == 'S') && (b == 'N')) return 9.0;
  if ((a == 'S') && (b == 'P')) return 9.0;
  if ((a == 'S') && (b == 'Q')) return 6.0;
  if ((a == 'S') && (b == 'R')) return 7.0;
  if ((a == 'S') && (b == 'S')) return 11.0;

  if ((a == 'T') && (b == 'A')) return 9.0;
  if ((a == 'T') && (b == 'B')) return 8.0;
  if ((a == 'T') && (b == 'C')) return 5.0;
  if ((a == 'T') && (b == 'D')) return 7.0;
  if ((a == 'T') && (b == 'E')) return 6.0;
  if ((a == 'T') && (b == 'F')) return 4.0;
  if ((a == 'T') && (b == 'G')) return 7.0;
  if ((a == 'T') && (b == 'H')) return 5.0;
  if ((a == 'T') && (b == 'I')) return 8.0;
  if ((a == 'T') && (b == 'K')) return 7.0;
  if ((a == 'T') && (b == 'L')) return 5.0;
  if ((a == 'T') && (b == 'M')) return 7.0;
  if ((a == 'T') && (b == 'N')) return 8.0;
  if ((a == 'T') && (b == 'P')) return 7.0;
  if ((a == 'T') && (b == 'Q')) return 6.0;
  if ((a == 'T') && (b == 'R')) return 6.0;
  if ((a == 'T') && (b == 'S')) return 10.0;
  if ((a == 'T') && (b == 'T')) return 12.0;

  if ((a == 'V') && (b == 'A')) return 8.0;
  if ((a == 'V') && (b == 'B')) return 5.0;
  if ((a == 'V') && (b == 'C')) return 5.0;
  if ((a == 'V') && (b == 'D')) return 5.0;
  if ((a == 'V') && (b == 'E')) return 5.0;
  if ((a == 'V') && (b == 'F')) return 5.0;
  if ((a == 'V') && (b == 'G')) return 6.0;
  if ((a == 'V') && (b == 'H')) return 5.0;
  if ((a == 'V') && (b == 'I')) return 11.0;
  if ((a == 'V') && (b == 'K')) return 4.0;
  if ((a == 'V') && (b == 'L')) return 9.0;
  if ((a == 'V') && (b == 'M')) return 9.0;
  if ((a == 'V') && (b == 'N')) return 5.0;
  if ((a == 'V') && (b == 'P')) return 6.0;
  if ((a == 'V') && (b == 'Q')) return 5.0;
  if ((a == 'V') && (b == 'R')) return 5.0;
  if ((a == 'V') && (b == 'S')) return 6.0;
  if ((a == 'V') && (b == 'T')) return 8.0;
  if ((a == 'V') && (b == 'V')) return 13.0;

  if ((a == 'W') && (b == 'A')) return 1.0;
  if ((a == 'W') && (b == 'B')) return 2.0;
  if ((a == 'W') && (b == 'C')) return 0.0;
  if ((a == 'W') && (b == 'D')) return 0.0;
  if ((a == 'W') && (b == 'E')) return 0.0;
  if ((a == 'W') && (b == 'F')) return 7.0;
  if ((a == 'W') && (b == 'G')) return 0.0;
  if ((a == 'W') && (b == 'H')) return 5.0;
  if ((a == 'W') && (b == 'I')) return 2.0;
  if ((a == 'W') && (b == 'K')) return 3.0;
  if ((a == 'W') && (b == 'L')) return 5.0;
  if ((a == 'W') && (b == 'M')) return 2.0;
  if ((a == 'W') && (b == 'N')) return 4.0;
  if ((a == 'W') && (b == 'P')) return 1.0;
  if ((a == 'W') && (b == 'Q')) return 2.0;
  if ((a == 'W') && (b == 'R')) return 9.0;
  if ((a == 'W') && (b == 'S')) return 6.0;
  if ((a == 'W') && (b == 'T')) return 2.0;
  if ((a == 'W') && (b == 'V')) return 0.0;
  if ((a == 'W') && (b == 'W')) return 20.0;

/*   if ((a == 'X') && (b == 'A')) return 2; */
/*   if ((a == 'X') && (b == 'B')) return 4; */
/*   if ((a == 'X') && (b == 'C')) return 0; */
/*   if ((a == 'X') && (b == 'D')) return 0; */
/*   if ((a == 'X') && (b == 'E')) return 0; */
/*   if ((a == 'X') && (b == 'F')) return 2; */
/*   if ((a == 'X') && (b == 'G')) return 0; */
/*   if ((a == 'X') && (b == 'H')) return 8; */
/*   if ((a == 'X') && (b == 'I')) return 4; */
/*   if ((a == 'X') && (b == 'K')) return 2; */
/*   if ((a == 'X') && (b == 'L')) return 2; */
/*   if ((a == 'X') && (b == 'M')) return 4; */
/*   if ((a == 'X') && (b == 'N')) return 6; */
/*   if ((a == 'X') && (b == 'P')) return 2; */
/*   if ((a == 'X') && (b == 'Q')) return 2; */
/*   if ((a == 'X') && (b == 'R')) return 6; */
/*   if ((a == 'X') && (b == 'S')) return 8; */
/*   if ((a == 'X') && (b == 'T')) return 4; */
/*   if ((a == 'X') && (b == 'V')) return 0; */
/*   if ((a == 'X') && (b == 'W')) return 0; */
/*   if ((a == 'X') && (b == 'X')) return 12; */

  if ((a == 'Y') && (b == 'A')) return 4.0;
  if ((a == 'Y') && (b == 'B')) return 5.0;
  if ((a == 'Y') && (b == 'C')) return 7.0;
  if ((a == 'Y') && (b == 'D')) return 3.0;
  if ((a == 'Y') && (b == 'E')) return 3.0;
  if ((a == 'Y') && (b == 'F')) return 12.0;
  if ((a == 'Y') && (b == 'G')) return 2.0;
  if ((a == 'Y') && (b == 'H')) return 7.0;
  if ((a == 'Y') && (b == 'I')) return 6.0;
  if ((a == 'Y') && (b == 'K')) return 3.0;
  if ((a == 'Y') && (b == 'L')) return 6.0;
  if ((a == 'Y') && (b == 'M')) return 4.0;
  if ((a == 'Y') && (b == 'N')) return 6.0;
  if ((a == 'Y') && (b == 'P')) return 2.0;
  if ((a == 'Y') && (b == 'Q')) return 3.0;
  if ((a == 'Y') && (b == 'R')) return 3.0;
  if ((a == 'Y') && (b == 'S')) return 5.0;
  if ((a == 'Y') && (b == 'T')) return 5.0;
  if ((a == 'Y') && (b == 'V')) return 5.0;
  if ((a == 'Y') && (b == 'W')) return 6.0;
/*   if ((a == 'Y') && (b == 'X')) return 4; */
  if ((a == 'Y') && (b == 'Y')) return 16.0;

  if ((a == 'Z') && (b == 'A')) return 7.0;
  if ((a == 'Z') && (b == 'B')) return 10.0;
  if ((a == 'Z') && (b == 'C')) return 1.0;
  if ((a == 'Z') && (b == 'D')) return 11.0;
  if ((a == 'Z') && (b == 'E')) return 12.0;
  if ((a == 'Z') && (b == 'F')) return 2.0;
  if ((a == 'Z') && (b == 'G')) return 6.0;
  if ((a == 'Z') && (b == 'H')) return 9.0;
  if ((a == 'Z') && (b == 'I')) return 5.0;
  if ((a == 'Z') && (b == 'K')) return 7.0;
  if ((a == 'Z') && (b == 'L')) return 5.0;
  if ((a == 'Z') && (b == 'M')) return 6.0;
  if ((a == 'Z') && (b == 'N')) return 8.0;
  if ((a == 'Z') && (b == 'P')) return 7.0;
  if ((a == 'Z') && (b == 'Q')) return 12.0;
  if ((a == 'Z') && (b == 'R')) return 7.0;
  if ((a == 'Z') && (b == 'S')) return 7.0;
  if ((a == 'Z') && (b == 'T')) return 6.0;
  if ((a == 'Z') && (b == 'V')) return 5.0;
  if ((a == 'Z') && (b == 'W')) return 1.0;
/*   if ((a == 'Z') && (b == 'X')) return 2; */
  if ((a == 'Z') && (b == 'Y')) return 3.0;
  if ((a == 'Z') && (b == 'Z')) return 12.0;

  if (a == b)
    return 6.0;
  else
    return 0.0;
} /* pam120_p */

/* citeseer.ist.psu.edu/naor96amino.html - Amino Acid Pair Interchanges at
 * Spatially Conserved Locations - Nussinov
 */

double nussinov(char a, char b, int altered, int use_carbonyl) {

  if (use_carbonyl) {
    if (isascii(a) != isascii(b))
      return -1000;
    else {
      if (! isascii(a)) {
	a = toascii(a);
	b = toascii(b);
	if (isalpha(a) != isalpha(b))
	  return -1000;
      }
      if (isalpha(a) && isalpha(b)) {
	if (isupper(a) != isupper(b))
	  return -1000;
	else if (islower(a)) {
	  a = _toupper(a);
	  b = _toupper(b);
	}
      }
    }
  }

  if (altered && (! (isalpha(a) && isalpha(b)))) {
    if (a == b)
      return 1000;
    else
      return -1000;
  }

  if (a < b) return nussinov(b,a,0,0);

  if ((a == 'A') && (b == 'A')) return 13;
  
  if ((a == 'B') && (b == 'A')) return 9;
  if ((a == 'B') && (b == 'B')) return 18.5;
  
  if ((a == 'C') && (b == 'A')) return 11;
  if ((a == 'C') && (b == 'B')) return 4.5;
  if ((a == 'C') && (b == 'C')) return 23;
  
  if ((a == 'D') && (b == 'A')) return 9;
  if ((a == 'D') && (b == 'B')) return 18.5;
  if ((a == 'D') && (b == 'C')) return 5;
  if ((a == 'D') && (b == 'D')) return 20;
  
  if ((a == 'E') && (b == 'A')) return 9;
  if ((a == 'E') && (b == 'B')) return 15.5;
  if ((a == 'E') && (b == 'C')) return 3;
  if ((a == 'E') && (b == 'D')) return 16;
  if ((a == 'E') && (b == 'E')) return 20;
  
  if ((a == 'F') && (b == 'A')) return 8;
  if ((a == 'F') && (b == 'B')) return 3;
  if ((a == 'F') && (b == 'C')) return 12;
  if ((a == 'F') && (b == 'D')) return 3;
  if ((a == 'F') && (b == 'E')) return 3;
  if ((a == 'F') && (b == 'F')) return 15;
  
  if ((a == 'G') && (b == 'A')) return 9;
  if ((a == 'G') && (b == 'B')) return 13.5;
  if ((a == 'G') && (b == 'C')) return 9;
  if ((a == 'G') && (b == 'D')) return 13;
  if ((a == 'G') && (b == 'E')) return 9;
  if ((a == 'G') && (b == 'F')) return 6;
  if ((a == 'G') && (b == 'G')) return 21;
  
  if ((a == 'H') && (b == 'A')) return 8;
  if ((a == 'H') && (b == 'B')) return 11.5;
  if ((a == 'H') && (b == 'C')) return 7;
  if ((a == 'H') && (b == 'D')) return 12;
  if ((a == 'H') && (b == 'E')) return 11;
  if ((a == 'H') && (b == 'F')) return 7;
  if ((a == 'H') && (b == 'G')) return 12;
  if ((a == 'H') && (b == 'H')) return 14;
  
  if ((a == 'I') && (b == 'A')) return 7;
  if ((a == 'I') && (b == 'B')) return 0;
  if ((a == 'I') && (b == 'C')) return 13;
  if ((a == 'I') && (b == 'D')) return 0;
  if ((a == 'I') && (b == 'E')) return 0;
  if ((a == 'I') && (b == 'F')) return 15;
  if ((a == 'I') && (b == 'G')) return 2;
  if ((a == 'I') && (b == 'H')) return 6;
  if ((a == 'I') && (b == 'I')) return 19;

  if ((a == 'K') && (b == 'A')) return 9;
  if ((a == 'K') && (b == 'B')) return 15.5;
  if ((a == 'K') && (b == 'C')) return 2;
  if ((a == 'K') && (b == 'D')) return 16;
  if ((a == 'K') && (b == 'E')) return 18;
  if ((a == 'K') && (b == 'F')) return 2;
  if ((a == 'K') && (b == 'G')) return 9;
  if ((a == 'K') && (b == 'H')) return 10;
  if ((a == 'K') && (b == 'I')) return 0;
  if ((a == 'K') && (b == 'K')) return 22;

  if ((a == 'L') && (b == 'A')) return 9;
  if ((a == 'L') && (b == 'B')) return 1.5;
  if ((a == 'L') && (b == 'C')) return 11;
  if ((a == 'L') && (b == 'D')) return 1;
  if ((a == 'L') && (b == 'E')) return 3;
  if ((a == 'L') && (b == 'F')) return 15;
  if ((a == 'L') && (b == 'G')) return 3;
  if ((a == 'L') && (b == 'H')) return 6;
  if ((a == 'L') && (b == 'I')) return 16;
  if ((a == 'L') && (b == 'K')) return 2;
  if ((a == 'L') && (b == 'L')) return 18;

  if ((a == 'M') && (b == 'A')) return 9;
  if ((a == 'M') && (b == 'B')) return 5.5;
  if ((a == 'M') && (b == 'C')) return 11;
  if ((a == 'M') && (b == 'D')) return 5;
  if ((a == 'M') && (b == 'E')) return 7;
  if ((a == 'M') && (b == 'F')) return 12;
  if ((a == 'M') && (b == 'G')) return 6;
  if ((a == 'M') && (b == 'H')) return 8;
  if ((a == 'M') && (b == 'I')) return 12;
  if ((a == 'M') && (b == 'K')) return 5;
  if ((a == 'M') && (b == 'L')) return 14;
  if ((a == 'M') && (b == 'M')) return 13;

  if ((a == 'N') && (b == 'A')) return 9;
  if ((a == 'N') && (b == 'B')) return 17;
  if ((a == 'N') && (b == 'C')) return 4;
  if ((a == 'N') && (b == 'D')) return 17;
  if ((a == 'N') && (b == 'E')) return 15;
  if ((a == 'N') && (b == 'F')) return 3;
  if ((a == 'N') && (b == 'G')) return 14;
  if ((a == 'N') && (b == 'H')) return 11;
  if ((a == 'N') && (b == 'I')) return 0;
  if ((a == 'N') && (b == 'K')) return 15;
  if ((a == 'N') && (b == 'L')) return 2;
  if ((a == 'N') && (b == 'M')) return 6;
  if ((a == 'N') && (b == 'N')) return 17;

  if ((a == 'P') && (b == 'A')) return 9;
  if ((a == 'P') && (b == 'B')) return 15;
  if ((a == 'P') && (b == 'C')) return 8;
  if ((a == 'P') && (b == 'D')) return 15;
  if ((a == 'P') && (b == 'E')) return 9;
  if ((a == 'P') && (b == 'F')) return 5;
  if ((a == 'P') && (b == 'G')) return 15;
  if ((a == 'P') && (b == 'H')) return 11;
  if ((a == 'P') && (b == 'I')) return 1;
  if ((a == 'P') && (b == 'K')) return 12;
  if ((a == 'P') && (b == 'L')) return 1;
  if ((a == 'P') && (b == 'M')) return 3;
  if ((a == 'P') && (b == 'N')) return 15;
  if ((a == 'P') && (b == 'P')) return 28;

  if ((a == 'Q') && (b == 'A')) return 9;
  if ((a == 'Q') && (b == 'B')) return 14.5;
  if ((a == 'Q') && (b == 'C')) return 5;
  if ((a == 'Q') && (b == 'D')) return 15;
  if ((a == 'Q') && (b == 'E')) return 17;
  if ((a == 'Q') && (b == 'F')) return 4;
  if ((a == 'Q') && (b == 'G')) return 9;
  if ((a == 'Q') && (b == 'H')) return 10;
  if ((a == 'Q') && (b == 'I')) return 2;
  if ((a == 'Q') && (b == 'K')) return 16;
  if ((a == 'Q') && (b == 'L')) return 4;
  if ((a == 'Q') && (b == 'M')) return 7;
  if ((a == 'Q') && (b == 'N')) return 14;
  if ((a == 'Q') && (b == 'P')) return 10;
  if ((a == 'Q') && (b == 'Q')) return 16;

  if ((a == 'R') && (b == 'A')) return 9;
  if ((a == 'R') && (b == 'B')) return 13;
  if ((a == 'R') && (b == 'C')) return 4;
  if ((a == 'R') && (b == 'D')) return 13;
  if ((a == 'R') && (b == 'E')) return 17;
  if ((a == 'R') && (b == 'F')) return 4;
  if ((a == 'R') && (b == 'G')) return 9;
  if ((a == 'R') && (b == 'H')) return 12;
  if ((a == 'R') && (b == 'I')) return 2;
  if ((a == 'R') && (b == 'K')) return 18;
  if ((a == 'R') && (b == 'L')) return 4;
  if ((a == 'R') && (b == 'M')) return 7;
  if ((a == 'R') && (b == 'N')) return 13;
  if ((a == 'R') && (b == 'P')) return 10;
  if ((a == 'R') && (b == 'Q')) return 14;
  if ((a == 'R') && (b == 'R')) return 16;

  if ((a == 'S') && (b == 'A')) return 9;
  if ((a == 'S') && (b == 'B')) return 13;
  if ((a == 'S') && (b == 'C')) return 6;
  if ((a == 'S') && (b == 'D')) return 13;
  if ((a == 'S') && (b == 'E')) return 12;
  if ((a == 'S') && (b == 'F')) return 5;
  if ((a == 'S') && (b == 'G')) return 13;
  if ((a == 'S') && (b == 'H')) return 11;
  if ((a == 'S') && (b == 'I')) return 4;
  if ((a == 'S') && (b == 'K')) return 12;
  if ((a == 'S') && (b == 'L')) return 4;
  if ((a == 'S') && (b == 'M')) return 6;
  if ((a == 'S') && (b == 'N')) return 13;
  if ((a == 'S') && (b == 'P')) return 12;
  if ((a == 'S') && (b == 'Q')) return 11;
  if ((a == 'S') && (b == 'R')) return 11;
  if ((a == 'S') && (b == 'S')) return 14;

  if ((a == 'T') && (b == 'A')) return 9;
  if ((a == 'T') && (b == 'B')) return 11;
  if ((a == 'T') && (b == 'C')) return 7;
  if ((a == 'T') && (b == 'D')) return 11;
  if ((a == 'T') && (b == 'E')) return 12;
  if ((a == 'T') && (b == 'F')) return 7;
  if ((a == 'T') && (b == 'G')) return 9;
  if ((a == 'T') && (b == 'H')) return 9;
  if ((a == 'T') && (b == 'I')) return 6;
  if ((a == 'T') && (b == 'K')) return 11;
  if ((a == 'T') && (b == 'L')) return 6;
  if ((a == 'T') && (b == 'M')) return 7;
  if ((a == 'T') && (b == 'N')) return 11;
  if ((a == 'T') && (b == 'P')) return 10;
  if ((a == 'T') && (b == 'Q')) return 10;
  if ((a == 'T') && (b == 'R')) return 10;
  if ((a == 'T') && (b == 'S')) return 11;
  if ((a == 'T') && (b == 'T')) return 12;

  if ((a == 'V') && (b == 'A')) return 7;
  if ((a == 'V') && (b == 'B')) return 0.5;
  if ((a == 'V') && (b == 'C')) return 13;
  if ((a == 'V') && (b == 'D')) return 0;
  if ((a == 'V') && (b == 'E')) return 0;
  if ((a == 'V') && (b == 'F')) return 14;
  if ((a == 'V') && (b == 'G')) return 4;
  if ((a == 'V') && (b == 'H')) return 7;
  if ((a == 'V') && (b == 'I')) return 18;
  if ((a == 'V') && (b == 'K')) return 1;
  if ((a == 'V') && (b == 'L')) return 14;
  if ((a == 'V') && (b == 'M')) return 11;
  if ((a == 'V') && (b == 'N')) return 1;
  if ((a == 'V') && (b == 'P')) return 4;
  if ((a == 'V') && (b == 'Q')) return 1;
  if ((a == 'V') && (b == 'R')) return 3;
  if ((a == 'V') && (b == 'S')) return 5;
  if ((a == 'V') && (b == 'T')) return 8;
  if ((a == 'V') && (b == 'V')) return 18;

  if ((a == 'W') && (b == 'A')) return 9;
  if ((a == 'W') && (b == 'B')) return 5;
  if ((a == 'W') && (b == 'C')) return 9;
  if ((a == 'W') && (b == 'D')) return 5;
  if ((a == 'W') && (b == 'E')) return 6;
  if ((a == 'W') && (b == 'F')) return 12;
  if ((a == 'W') && (b == 'G')) return 8;
  if ((a == 'W') && (b == 'H')) return 8;
  if ((a == 'W') && (b == 'I')) return 13;
  if ((a == 'W') && (b == 'K')) return 4;
  if ((a == 'W') && (b == 'L')) return 11;
  if ((a == 'W') && (b == 'M')) return 10;
  if ((a == 'W') && (b == 'N')) return 5;
  if ((a == 'W') && (b == 'P')) return 4;
  if ((a == 'W') && (b == 'Q')) return 7;
  if ((a == 'W') && (b == 'R')) return 6;
  if ((a == 'W') && (b == 'S')) return 7;
  if ((a == 'W') && (b == 'T')) return 9;
  if ((a == 'W') && (b == 'V')) return 13;
  if ((a == 'W') && (b == 'W')) return 18;

  if ((a == 'Y') && (b == 'A')) return 9;
  if ((a == 'Y') && (b == 'B')) return 6.5;
  if ((a == 'Y') && (b == 'C')) return 9;
  if ((a == 'Y') && (b == 'D')) return 6;
  if ((a == 'Y') && (b == 'E')) return 7;
  if ((a == 'Y') && (b == 'F')) return 11;
  if ((a == 'Y') && (b == 'G')) return 8;
  if ((a == 'Y') && (b == 'H')) return 9;
  if ((a == 'Y') && (b == 'I')) return 11;
  if ((a == 'Y') && (b == 'K')) return 6;
  if ((a == 'Y') && (b == 'L')) return 10;
  if ((a == 'Y') && (b == 'M')) return 10;
  if ((a == 'Y') && (b == 'N')) return 7;
  if ((a == 'Y') && (b == 'P')) return 6;
  if ((a == 'Y') && (b == 'Q')) return 8;
  if ((a == 'Y') && (b == 'R')) return 7;
  if ((a == 'Y') && (b == 'S')) return 8;
  if ((a == 'Y') && (b == 'T')) return 9;
  if ((a == 'Y') && (b == 'V')) return 11;
  if ((a == 'Y') && (b == 'W')) return 11;
  if ((a == 'Y') && (b == 'Y')) return 12;

  if ((a == 'Z') && (b == 'A')) return 9;
  if ((a == 'Z') && (b == 'B')) return 15;
  if ((a == 'Z') && (b == 'C')) return 4;
  if ((a == 'Z') && (b == 'D')) return 15.5;
  if ((a == 'Z') && (b == 'E')) return 18.5;
  if ((a == 'Z') && (b == 'F')) return 3.5;
  if ((a == 'Z') && (b == 'G')) return 9;
  if ((a == 'Z') && (b == 'H')) return 10.5;
  if ((a == 'Z') && (b == 'I')) return 1;
  if ((a == 'Z') && (b == 'K')) return 17;
  if ((a == 'Z') && (b == 'L')) return 3.5;
  if ((a == 'Z') && (b == 'M')) return 7;
  if ((a == 'Z') && (b == 'N')) return 14.5;
  if ((a == 'Z') && (b == 'P')) return 9.5;
  if ((a == 'Z') && (b == 'Q')) return 16.5;
  if ((a == 'Z') && (b == 'R')) return 15.5;
  if ((a == 'Z') && (b == 'S')) return 11.5;
  if ((a == 'Z') && (b == 'T')) return 11;
  if ((a == 'Z') && (b == 'V')) return 0.5;
  if ((a == 'Z') && (b == 'W')) return 6.5;
  if ((a == 'Z') && (b == 'Y')) return 7.5;
  if ((a == 'Z') && (b == 'Z')) return 17;

  if (a == b)
    return 12;
  else
    return 0;
} /* nussinov */

double blosum62(char a, char b, int altered, int use_carbonyl) {

  if (use_carbonyl) {
    if (isascii(a) != isascii(b))
      return -1000;
    else {
      if (! isascii(a)) {
	a = toascii(a);
	b = toascii(b);
	if (isalpha(a) != isalpha(b))
	  return -1000;
      }
      if (isalpha(a) && isalpha(b)) {
	if (isupper(a) != isupper(b))
	  return -1000;
	else if (islower(a)) {
	  a = _toupper(a);
	  b = _toupper(b);
	}
      }
    }
  }

  if (altered && (! (isalpha(a) && isalpha(b)))) {
    if (a == b)
      return 1000;
    else
      return -1000;
  }

  if (a < b) return blosum62(b,a,0,0);

  if ( a == 'A' && b == 'A' ) return    7;
  
  if ( a == 'B' && b == 'A' ) return   -4;
  if ( a == 'B' && b == 'B' ) return    7;

  if ( a == 'C' && b == 'A' ) return    0;
  if ( a == 'C' && b == 'B' ) return   -6;
  if ( a == 'C' && b == 'C' ) return   17;

  if ( a == 'D' && b == 'A' ) return   -4;
  if ( a == 'D' && b == 'B' ) return    8;
  if ( a == 'D' && b == 'C' ) return   -6;
  if ( a == 'D' && b == 'D' ) return   11;

  if ( a == 'E' && b == 'A' ) return   -2;
  if ( a == 'E' && b == 'B' ) return    2;
  if ( a == 'E' && b == 'C' ) return   -8;
  if ( a == 'E' && b == 'D' ) return    4;
  if ( a == 'E' && b == 'E' ) return    9;

  if ( a == 'F' && b == 'A' ) return   -4;
  if ( a == 'F' && b == 'B' ) return   -6;
  if ( a == 'F' && b == 'C' ) return   -4;
  if ( a == 'F' && b == 'D' ) return   -6;
  if ( a == 'F' && b == 'E' ) return   -6;
  if ( a == 'F' && b == 'F' ) return   11;

  if ( a == 'G' && b == 'A' ) return    0;
  if ( a == 'G' && b == 'B' ) return   -2;
  if ( a == 'G' && b == 'C' ) return   -6;
  if ( a == 'G' && b == 'D' ) return   -2;
  if ( a == 'G' && b == 'E' ) return   -4;
  if ( a == 'G' && b == 'F' ) return   -6;
  if ( a == 'G' && b == 'G' ) return   11;

  if ( a == 'H' && b == 'A' ) return   -4;
  if ( a == 'H' && b == 'B' ) return    0;
  if ( a == 'H' && b == 'C' ) return   -6;
  if ( a == 'H' && b == 'D' ) return   -2;
  if ( a == 'H' && b == 'E' ) return    0;
  if ( a == 'H' && b == 'F' ) return   -2;
  if ( a == 'H' && b == 'G' ) return   -4;
  if ( a == 'H' && b == 'H' ) return   15;

  if ( a == 'I' && b == 'A' ) return   -2;
  if ( a == 'I' && b == 'B' ) return   -6;
  if ( a == 'I' && b == 'C' ) return   -2;
  if ( a == 'I' && b == 'D' ) return   -6;
  if ( a == 'I' && b == 'E' ) return   -6;
  if ( a == 'I' && b == 'F' ) return    0;
  if ( a == 'I' && b == 'G' ) return   -8;
  if ( a == 'I' && b == 'H' ) return   -6;
  if ( a == 'I' && b == 'I' ) return    7;

  if ( a == 'K' && b == 'A' ) return   -2;
  if ( a == 'K' && b == 'B' ) return    0;
  if ( a == 'K' && b == 'C' ) return   -6;
  if ( a == 'K' && b == 'D' ) return   -2;
  if ( a == 'K' && b == 'E' ) return    2;
  if ( a == 'K' && b == 'F' ) return   -6;
  if ( a == 'K' && b == 'G' ) return   -4;
  if ( a == 'K' && b == 'H' ) return   -2;
  if ( a == 'K' && b == 'I' ) return   -6;
  if ( a == 'K' && b == 'K' ) return    9;

  if ( a == 'L' && b == 'A' ) return   -2;
  if ( a == 'L' && b == 'B' ) return   -8;
  if ( a == 'L' && b == 'C' ) return   -2;
  if ( a == 'L' && b == 'D' ) return   -8;
  if ( a == 'L' && b == 'E' ) return   -6;
  if ( a == 'L' && b == 'F' ) return    0;
  if ( a == 'L' && b == 'G' ) return   -8;
  if ( a == 'L' && b == 'H' ) return   -6;
  if ( a == 'L' && b == 'I' ) return    4;
  if ( a == 'L' && b == 'K' ) return   -4;
  if ( a == 'L' && b == 'L' ) return    7;

  if ( a == 'M' && b == 'A' ) return   -2;
  if ( a == 'M' && b == 'B' ) return   -6;
  if ( a == 'M' && b == 'C' ) return   -2;
  if ( a == 'M' && b == 'D' ) return   -6;
  if ( a == 'M' && b == 'E' ) return   -4;
  if ( a == 'M' && b == 'F' ) return    0;
  if ( a == 'M' && b == 'G' ) return   -6;
  if ( a == 'M' && b == 'H' ) return   -4;
  if ( a == 'M' && b == 'I' ) return    2;
  if ( a == 'M' && b == 'K' ) return   -2;
  if ( a == 'M' && b == 'L' ) return    4;
  if ( a == 'M' && b == 'M' ) return    9;

  if ( a == 'N' && b == 'A' ) return   -4;
  if ( a == 'N' && b == 'B' ) return    6;
  if ( a == 'N' && b == 'C' ) return   -6;
  if ( a == 'N' && b == 'D' ) return    2;
  if ( a == 'N' && b == 'E' ) return    0;
  if ( a == 'N' && b == 'F' ) return   -6;
  if ( a == 'N' && b == 'G' ) return    0;
  if ( a == 'N' && b == 'H' ) return    2;
  if ( a == 'N' && b == 'I' ) return   -6;
  if ( a == 'N' && b == 'K' ) return    0;
  if ( a == 'N' && b == 'L' ) return   -6;
  if ( a == 'N' && b == 'M' ) return   -4;
  if ( a == 'N' && b == 'N' ) return   11;

  if ( a == 'P' && b == 'A' ) return   -2;
  if ( a == 'P' && b == 'B' ) return   -4;
  if ( a == 'P' && b == 'C' ) return   -6;
  if ( a == 'P' && b == 'D' ) return   -2;
  if ( a == 'P' && b == 'E' ) return   -2;
  if ( a == 'P' && b == 'F' ) return   -8;
  if ( a == 'P' && b == 'G' ) return   -4;
  if ( a == 'P' && b == 'H' ) return   -4;
  if ( a == 'P' && b == 'I' ) return   -6;
  if ( a == 'P' && b == 'K' ) return   -2;
  if ( a == 'P' && b == 'L' ) return   -6;
  if ( a == 'P' && b == 'M' ) return   -4;
  if ( a == 'P' && b == 'N' ) return   -4;
  if ( a == 'P' && b == 'P' ) return   13;

  if ( a == 'Q' && b == 'A' ) return   -2;
  if ( a == 'Q' && b == 'B' ) return    0;
  if ( a == 'Q' && b == 'C' ) return   -6;
  if ( a == 'Q' && b == 'D' ) return    0;
  if ( a == 'Q' && b == 'E' ) return    4;
  if ( a == 'Q' && b == 'F' ) return   -6;
  if ( a == 'Q' && b == 'G' ) return   -4;
  if ( a == 'Q' && b == 'H' ) return    0;
  if ( a == 'Q' && b == 'I' ) return   -6;
  if ( a == 'Q' && b == 'K' ) return    2;
  if ( a == 'Q' && b == 'L' ) return   -4;
  if ( a == 'Q' && b == 'M' ) return    0;
  if ( a == 'Q' && b == 'N' ) return    0;
  if ( a == 'Q' && b == 'P' ) return   -2;
  if ( a == 'Q' && b == 'Q' ) return    9;

  if ( a == 'R' && b == 'A' ) return   -2;
  if ( a == 'R' && b == 'B' ) return   -2;
  if ( a == 'R' && b == 'C' ) return   -6;
  if ( a == 'R' && b == 'D' ) return   -4;
  if ( a == 'R' && b == 'E' ) return    0;
  if ( a == 'R' && b == 'F' ) return   -6;
  if ( a == 'R' && b == 'G' ) return   -4;
  if ( a == 'R' && b == 'H' ) return    0;
  if ( a == 'R' && b == 'I' ) return   -6;
  if ( a == 'R' && b == 'K' ) return    4;
  if ( a == 'R' && b == 'L' ) return   -4;
  if ( a == 'R' && b == 'M' ) return   -2;
  if ( a == 'R' && b == 'N' ) return    0;
  if ( a == 'R' && b == 'P' ) return   -4;
  if ( a == 'R' && b == 'Q' ) return    2;
  if ( a == 'R' && b == 'R' ) return    9;

  if ( a == 'S' && b == 'A' ) return    2;
  if ( a == 'S' && b == 'B' ) return    0;
  if ( a == 'S' && b == 'C' ) return   -2;
  if ( a == 'S' && b == 'D' ) return    0;
  if ( a == 'S' && b == 'E' ) return    0;
  if ( a == 'S' && b == 'F' ) return   -4;
  if ( a == 'S' && b == 'G' ) return    0;
  if ( a == 'S' && b == 'H' ) return   -2;
  if ( a == 'S' && b == 'I' ) return   -4;
  if ( a == 'S' && b == 'K' ) return    0;
  if ( a == 'S' && b == 'L' ) return   -4;
  if ( a == 'S' && b == 'M' ) return   -2;
  if ( a == 'S' && b == 'N' ) return    2;
  if ( a == 'S' && b == 'P' ) return   -2;
  if ( a == 'S' && b == 'Q' ) return    0;
  if ( a == 'S' && b == 'R' ) return   -2;
  if ( a == 'S' && b == 'S' ) return    7;

  if ( a == 'T' && b == 'A' ) return    0;
  if ( a == 'T' && b == 'B' ) return   -2;
  if ( a == 'T' && b == 'C' ) return   -2;
  if ( a == 'T' && b == 'D' ) return   -2;
  if ( a == 'T' && b == 'E' ) return   -2;
  if ( a == 'T' && b == 'F' ) return   -4;
  if ( a == 'T' && b == 'G' ) return   -4;
  if ( a == 'T' && b == 'H' ) return   -4;
  if ( a == 'T' && b == 'I' ) return   -2;
  if ( a == 'T' && b == 'K' ) return   -2;
  if ( a == 'T' && b == 'L' ) return   -2;
  if ( a == 'T' && b == 'M' ) return   -2;
  if ( a == 'T' && b == 'N' ) return    0;
  if ( a == 'T' && b == 'P' ) return   -2;
  if ( a == 'T' && b == 'Q' ) return   -2;
  if ( a == 'T' && b == 'R' ) return   -2;
  if ( a == 'T' && b == 'S' ) return    2;
  if ( a == 'T' && b == 'T' ) return    9;

  if ( a == 'V' && b == 'A' ) return    0;
  if ( a == 'V' && b == 'B' ) return   -6;
  if ( a == 'V' && b == 'C' ) return   -2;
  if ( a == 'V' && b == 'D' ) return   -6;
  if ( a == 'V' && b == 'E' ) return   -4;
  if ( a == 'V' && b == 'F' ) return   -2;
  if ( a == 'V' && b == 'G' ) return   -6;
  if ( a == 'V' && b == 'H' ) return   -6;
  if ( a == 'V' && b == 'I' ) return    6;
  if ( a == 'V' && b == 'K' ) return   -4;
  if ( a == 'V' && b == 'L' ) return    2;
  if ( a == 'V' && b == 'M' ) return    2;
  if ( a == 'V' && b == 'N' ) return   -6;
  if ( a == 'V' && b == 'P' ) return   -4;
  if ( a == 'V' && b == 'Q' ) return   -4;
  if ( a == 'V' && b == 'R' ) return   -6;
  if ( a == 'V' && b == 'S' ) return   -4;
  if ( a == 'V' && b == 'T' ) return    0;
  if ( a == 'V' && b == 'V' ) return    7;

  if ( a == 'W' && b == 'A' ) return   -6;
  if ( a == 'W' && b == 'B' ) return   -8;
  if ( a == 'W' && b == 'C' ) return   -4;
  if ( a == 'W' && b == 'D' ) return   -8;
  if ( a == 'W' && b == 'E' ) return   -6;
  if ( a == 'W' && b == 'F' ) return    2;
  if ( a == 'W' && b == 'G' ) return   -4;
  if ( a == 'W' && b == 'H' ) return   -4;
  if ( a == 'W' && b == 'I' ) return   -6;
  if ( a == 'W' && b == 'K' ) return   -6;
  if ( a == 'W' && b == 'L' ) return   -4;
  if ( a == 'W' && b == 'M' ) return   -2;
  if ( a == 'W' && b == 'N' ) return   -8;
  if ( a == 'W' && b == 'P' ) return   -8;
  if ( a == 'W' && b == 'Q' ) return   -4;
  if ( a == 'W' && b == 'R' ) return   -6;
  if ( a == 'W' && b == 'S' ) return   -6;
  if ( a == 'W' && b == 'T' ) return   -4;
  if ( a == 'W' && b == 'V' ) return   -6;
  if ( a == 'W' && b == 'W' ) return   21;

  if ( a == 'X' && b == 'A' ) return    0;
  if ( a == 'X' && b == 'B' ) return   -2;
  if ( a == 'X' && b == 'C' ) return   -4;
  if ( a == 'X' && b == 'D' ) return   -2;
  if ( a == 'X' && b == 'E' ) return   -2;
  if ( a == 'X' && b == 'F' ) return   -2;
  if ( a == 'X' && b == 'G' ) return   -2;
  if ( a == 'X' && b == 'H' ) return   -2;
  if ( a == 'X' && b == 'I' ) return   -2;
  if ( a == 'X' && b == 'K' ) return   -2;
  if ( a == 'X' && b == 'L' ) return   -2;
  if ( a == 'X' && b == 'M' ) return   -2;
  if ( a == 'X' && b == 'N' ) return   -2;
  if ( a == 'X' && b == 'P' ) return   -4;
  if ( a == 'X' && b == 'Q' ) return   -2;
  if ( a == 'X' && b == 'R' ) return   -2;
  if ( a == 'X' && b == 'S' ) return    0;
  if ( a == 'X' && b == 'T' ) return    0;
  if ( a == 'X' && b == 'V' ) return   -2;
  if ( a == 'X' && b == 'W' ) return   -4;
  if ( a == 'X' && b == 'X' ) return   -2;

  if ( a == 'Y' && b == 'A' ) return   -4;
  if ( a == 'Y' && b == 'B' ) return   -6;
  if ( a == 'Y' && b == 'C' ) return   -4;
  if ( a == 'Y' && b == 'D' ) return   -6;
  if ( a == 'Y' && b == 'E' ) return   -4;
  if ( a == 'Y' && b == 'F' ) return    6;
  if ( a == 'Y' && b == 'G' ) return   -6;
  if ( a == 'Y' && b == 'H' ) return    4;
  if ( a == 'Y' && b == 'I' ) return   -2;
  if ( a == 'Y' && b == 'K' ) return   -4;
  if ( a == 'Y' && b == 'L' ) return   -2;
  if ( a == 'Y' && b == 'M' ) return   -2;
  if ( a == 'Y' && b == 'N' ) return   -4;
  if ( a == 'Y' && b == 'P' ) return   -6;
  if ( a == 'Y' && b == 'Q' ) return   -2;
  if ( a == 'Y' && b == 'R' ) return   -4;
  if ( a == 'Y' && b == 'S' ) return   -4;
  if ( a == 'Y' && b == 'T' ) return   -4;
  if ( a == 'Y' && b == 'V' ) return   -2;
  if ( a == 'Y' && b == 'W' ) return    4;
  if ( a == 'Y' && b == 'X' ) return   -2;
  if ( a == 'Y' && b == 'Y' ) return   13;

  if ( a == 'Z' && b == 'A' ) return   -2;
  if ( a == 'Z' && b == 'B' ) return    2;
  if ( a == 'Z' && b == 'C' ) return   -6;
  if ( a == 'Z' && b == 'D' ) return    2;
  if ( a == 'Z' && b == 'E' ) return    8;
  if ( a == 'Z' && b == 'F' ) return   -6;
  if ( a == 'Z' && b == 'G' ) return   -4;
  if ( a == 'Z' && b == 'H' ) return    0;
  if ( a == 'Z' && b == 'I' ) return   -6;
  if ( a == 'Z' && b == 'K' ) return    2;
  if ( a == 'Z' && b == 'L' ) return   -6;
  if ( a == 'Z' && b == 'M' ) return   -2;
  if ( a == 'Z' && b == 'N' ) return    0;
  if ( a == 'Z' && b == 'P' ) return   -2;
  if ( a == 'Z' && b == 'Q' ) return    6;
  if ( a == 'Z' && b == 'R' ) return    0;
  if ( a == 'Z' && b == 'S' ) return    0;
  if ( a == 'Z' && b == 'T' ) return   -2;
  if ( a == 'Z' && b == 'V' ) return   -4;
  if ( a == 'Z' && b == 'W' ) return   -6;
  if ( a == 'Z' && b == 'X' ) return   -2;
  if ( a == 'Z' && b == 'Y' ) return   -4;
  if ( a == 'Z' && b == 'Z' ) return    7;

  if (a == b)
    return 0;
  else
    return -8;
} /* blosum62 */

double check_carbonyl(char a, char b) {
  if (isascii(a) != isascii(b))
    return -1000;
  else {
    if (! isascii(a)) {
      a = toascii(a);
      b = toascii(b);
      if (isalpha(a) != isalpha(b))
	return -1000;
    }
    if (isalpha(a) && isalpha(b)) {
      if (isupper(a) != isupper(b))
	return -1000;
    }
  }
  return 0;
}
  

/* Compromise between blosum62 and _really_ simpleminded version... - Allen */

double glyproaln(char a, char b, int altered, int use_carbonyl) {

  if (use_carbonyl) {
    if (isascii(a) != isascii(b))
      return -1000;
    else {
      if (! isascii(a)) {
	a = toascii(a);
	b = toascii(b);
	if (isalpha(a) != isalpha(b))
	  return -1000;
      }
      if (isalpha(a) && isalpha(b)) {
	if (isupper(a) != isupper(b))
	  return -1000;
	else if (islower(a)) {
	  a = _toupper(a);
	  b = _toupper(b);
	}
      }
    }
  }

  /* 1.0 for original (addition if identical),
     8.0 for conversion to positive */

  if (a == b) {
    if ((altered > 0) && (! isalpha(a)))
      return 1000;
    else if (a == 'X')
      return (1.0+8.0);
    else if (a == 'G')
      return (1.0+8.0+12.0);
    else if (a == 'P')
      return (1.0+8.0+14.0);
    else
      return (1.0+8.0+8.0);
  } else if (altered && (! (isalpha(a) && isalpha(b))))
    return -1000;
  else if ((a == 'G') || (b == 'G'))
    return 0.0;
  else if ((a == 'P') || (b == 'P'))
    return 0.0;
  else if (((a == 'B') && ((b == 'N') || (b == 'D'))) ||
	   ((b == 'B') && ((a == 'N') || (a == 'D'))))
    return (4.0+8.0);
  else if (((a == 'Z') && ((b == 'Q') || (b == 'E'))) ||
	   ((b == 'Z') && ((a == 'Q') || (a == 'E'))))
    return (4.0+8.0);
  else
    return 8.0;
} /* glyproaln */
